xquery version "3.1";
(:
 : Module with functions producing random distributions of various kinds.
 :
 : This version is XQuery 3.1 compliant. It is not as performant as the
 : MarkLogic version because random-number-generator() just isn't as usable as
 : xdmp:random. This works through the ugly hack of relying on generate-id() 
 : on a constructed node to produce new seeds. That isn't entirely random
 : because whenever Saxon decides that one expression is the same as another
 : it will optimize it to produce the same result, but for the most part
 : this works ok. With Saxon-PE or Saxon-EE I use timestamp() which is more
 : reliable.
 :
 : Even generating a sequence of numbers off random-number-generator isn't
 : that great because you need to construct maps to hold both the next
 : generator function and the number, or use recursion in a way highly
 : likely to generate stack overflows. (The example in the spec suffers this
 : problem.)
 :
 : The problem is that for my purposes I need to be able to create multiple
 : random sequences in the same query that are different from the sequences
 : in subsequent episodes. Where the sequences are partitioned into separate
 : modules so even if I did the ugly thing of making every function everywhere
 : have to cart around and return both the values I cared about and the 
 : current randomizer (which XQuery is so very not good at), and used explicit
 : distinct seeds: whence those seeds? They need to be different based on 
 : global program context and different from each run of the program. So
 : current-dateTime() concatenated with a value from a value range that is
 : preassigned to each component module. Or: cart around yet another randomizer
 : to provide the seeds. It is ugly, ugly, ugly any way you slice it.
 :
 : Usage:
 : You can either use the explicit distribution functions here, e.g.
 : rand:normal(10.0, 2.0), or construct a distribution descriptor using
 : the distributions API and then pass it into a call to rand:randomize(), 
 : e.g. rand:randomize(dist:normal(10.0, 2.0))
 : The latter approach allows for more fine control over the behaviour of
 : the final result (e.g. you can set a minimum value, a post-multiplier, etc.)
 : and is useful when you have a randomization algorithm that is getting
 : reused a lot, or that you might want to change up without having to find
 : all the places where it is called in the code.
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/core/random";

import module namespace dist="http://mathling.com/type/distribution"
       at "../types/distributions.xqy";
import module namespace errors="http://mathling.com/core/errors"
       at "errors.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

(:
 : Function to hand out seeds: 
 : Ugly hack alert: browbeat r-n-g into using a new seed => new number
 : See top comment. If you start getting samey-samey sequences all I 
 : can suggest is to spring for Saxon-PE.
 :)
declare variable $this:SEED as function() as xs:anyAtomicType :=
  let $timestamp := 
    function-lookup(QName("http://saxon.sf.net/", "timestamp"), 0)
  let $generate-id := 
    function-lookup(QName("http://www.w3.org/2005/xpath-functions", "generate-id"), 1)
  return
    if (exists($timestamp)) then $timestamp
    else (
      function () as xs:anyAtomicType {
        concat(current-dateTime(), $generate-id(<n/>))
      }
    )
;

(: 
 : ε is used to handle the switch from exclusive to inclusive bounds.
 :)
declare variable $this:SCALING as xs:integer := 100000;
declare %private variable $this:ε as xs:double := (1 div $this:SCALING) cast as xs:double;

(:
 : Number of fractional decimal digits to keep when we cast to decimal.
 :)
declare variable $this:DECIMAL-DIGITS as xs:integer := 2;
declare %private variable $this:DECIMAL-DIV as xs:integer := math:pow(10, $this:DECIMAL-DIGITS) cast as xs:integer;

declare variable $this:π as xs:double := math:pi();

(:
 : Number of resamplings will we try before we give up
 :)
declare variable $this:RESAMPLE-LIMIT as xs:integer := 10;

(: 
 : Uniform distribution [0,1.0) 
 :)
declare variable $this:STD-UNIFORM as map(xs:string,item()*) := map {
  "distribution" : "uniform",
  "min" : 0.0, 
  "max" : 1.0 - $this:ε
};

(: 
 : Normal distribution mean=0 std=1
 :)
declare variable $this:STD-NORMAL as map(xs:string,item()*) := map {
  "distribution" : "normal", 
  "mean" : 0.0, 
  "std" : 1.0
};

(: 
 : flip()
 : Returns random Bernoulli distributed data as a boolean
 :
 : $percent: Percent that should come up true()
 :)
declare function this:flip($percent as xs:double) as xs:boolean
{
  random-number-generator($this:SEED())=>map:get("number") lt $percent
};

(:
 : constant()
 : Returns a constant value, either the average of $min and $max (if they
 : exist) or whichever one does exist. Failing everything else, 1.
 :
 : This is the degenerate case of uniform() and exists as a convenience 
 : for data-driven randomization in randomize().
 : 
 : $min: minimum value
 : $max: maximum value
 :)
declare function this:constant($min as xs:double?, $max as xs:double?) as xs:double
{
  if (exists($min) and exists($max)) then (
    if ($max < $min) then errors:error("RAND-BADMINMAX",($min,$max))
    else if ($max = $min) then $min
    else ($max + $min) div 2
  ) else (
    ($min,$max,1)[1]
  )
};

(: 
 : uniform()
 : Return random uniformly distributed data in the range.
 : Because the base randomizer is exclusive on the upper bound, add ε
 : and then cap. This creates a slight over-weighting of the max value,
 : depending on fineness of ε.
 :
 : When we are using integers, we need to avoid chopping off the top of
 : the range, or we lose. We can get it, but it is unlikely and for small
 : integer ranges spectacularly unlikely. So we add 1-ε instead of ε and
 : quantize at this level. You want to avoid this and get, say, floating
 : point values between 5 and 10 (inc), pass floating point arguments.
 : 
 : Base randomizer gives [0,1) we want [x,y]
 :   [0,1) * (ε+y-x)  => [0,ε+y-x) 
 :   [0,ε+y-x) + x    => [x, y+ε)
 :   min(([x, y+ε),y) => [x,y]
 : 
 : For integers:
 :   [0,1) * (1-ε+y-x)  => [0,1-ε+y-x) 
 :   [0,1-ε+y-x) + x    => [x, y+1-ε)
 :   [x,y+1-ε) idiv 1   => [x, y]
 :
 : $min: Lower bound (inclusive) of range
 : $max: Upper bound (inclusive) of range
 :)
declare function this:uniform($min as xs:numeric, $max as xs:numeric) as xs:double
{
  if ($max < $min) then errors:error("RAND-BADMINMAX",($min,$max))
  else if ($max = $min) then $min 
  else if (
    ($max instance of xs:integer) and ($min instance of xs:integer)
  ) then (
    let $random_0_1 := random-number-generator($this:SEED())=>map:get("number")
    return ((1 - $this:ε + $max - $min)*$random_0_1 + $min) idiv 1
  ) else (
    let $random_0_1 := random-number-generator($this:SEED())=>map:get("number")
    return min(( ($this:ε + $max - $min)*$random_0_1 + $min, $max ))
  )
};

(: 
 : normal()
 : Return random normally distributed data.
 : 
 : $mean: Mean of range of values
 : $std: Standard deviation of range of values
 :)
declare function this:normal($mean as xs:double, $std as xs:double) as xs:double
{
  let $randomizer := random-number-generator($this:SEED())
  return $mean + this:gauss($randomizer) * $std
};

(: 
 : gauss()
 : Return random normally distributed data between 0 and 1
 : A service function used by normal() 
 :)
declare %private function this:gauss($randomizer as map(xs:string,item())) as xs:double
{
  let $u as xs:double := 2 * $randomizer=>map:get("number") - 1
  let $next := ($randomizer=>map:get("next"))()
  let $v as xs:double := 2 * $next=>map:get("number") - 1
  let $r := $u * $u + $v * $v
  return (
    if ($r = 0 or $r >= 1) then this:gauss(($next=>map:get("next"))())
    else (
      let $c as xs:double := math:sqrt(-2 * math:log($r) div $r)
      return $u * $c
    )
  )
};

(: 
 : skewed()
 : Return random data in skewed normal distribution.
 :
 : $mean: Mean of distribution
 : $std: Standard deviation of distribution
 : $skew: Skew of distribution
 : 
 : Skewed defined by parameters α, ξ, and ω
 : α is skew 
 :   α = 0 => normal, α > 0 right skewed, α < 0 left skewed
 :      right skewed => long tail on right
 : ξ = location = shift along x
 : ω = scaling along y
 :   ω is positive 
 : 
 : mean = ξ + ω*δ*sqrt(2/π) δ=α/sqrt(1+α*α) 
 : std = ω*ω*(1 - 2*δ*δ/π)
 :
 : So:
 : ω = sqrt(std / (1 - 2*δ*δ/π))
 : ξ = mean - ω*δ*sqrt(2/π)
 :)
declare function this:skewed($mean as xs:double, $std as xs:double, $skew as xs:double) as xs:double
{
  let $α := $skew
  let $δ := $α div math:sqrt(1 + $α*$α)
  let $ω := math:sqrt($std div (1 - (2*$δ*$δ div $this:π)))
  let $ξ := $mean - $ω*$δ*math:sqrt(2 div $this:π)
  let $σ := $α div math:sqrt(1 + $α*$α)
  let $u := this:normal(0, 1)
  let $v := this:normal(0, 1)
  let $u1 := $σ*$u + math:sqrt(1 - $σ*$σ)*$v
  return (
    if ($u lt 0)
    then -$ω * $u1 + $ξ
    else $ω * $u1 + $ξ
  )
};


(:
 : shuffle()
 : Return the data items in a (uniformly) random order.
 : 
 : $data: Data items to shuffle
 :)
declare function this:shuffle($data as item()*)
{
  random-number-generator($this:SEED())?permute($data)
};

(:
 : percent-sums()
 : Returns a sequence of cumulative probabilities, to be used
 : by select-index() to perform selections. This is to ensure we don't have to
 : keep recomputing the cumulative sums for each selection.
 :
 : To account for rounding, the last item is always forced to 1 and any sums
 : greater than 1 are also rounded down to 1. To get expected distributions, 
 : ensure that the input percentages sum to 100.
 :
 : Sums are returned in map:keys() order, so this assumes stability of that
 : function between calls to percent-sums() and calls to selection(). 
 : 
 : $weights a map from key to percent (expressed as integer 0 to 100)
 :)
declare function this:percent-sums($weights as map(xs:string,item()*)) as xs:double*
{
  let $percentages :=
    for $k in $weights=>map:keys() return $weights=>map:get($k)
  let $sums :=
    fold-left(
      $percentages, 0, 
      function($z,$a) {
        $z,
        min(($z[last()] + ($a div 100), 1))
      }
    )[position()=(2 to last())]
  return (
    if ($sums[last()] lt 1) then ($sums[position()=(1 to last()-1)],1) else $sums
  )
};

(:
 : zipf-sums()
 : Returns a sequence of cumulative Zipf probabilities, to be used
 : by select-index() to perform selections. This is to ensure we don't have to
 : keep recomputing the cumulative sums for each selection.
 : 
 : p(n)=(1/k^α)/sum(i=1 to n)(1/i^α)
 :    n = number of elements
 :    k = rank
 :    α = exponent
 :
 : $alpha: the alpha parameter, should be >=1
 :   For English words this by some estimates is around 1.6
 :   For city sizes 1.07
 : $n: number to generate
 :)
declare function this:zipf-sums($alpha as xs:double, $n as xs:integer) as xs:double*
{
   let $c := 
     fold-left(
       1 to $n, 0,
       function($z,$a) {
          $z + 1 div (math:pow(xs:double($a), $alpha))
       }
     )
   let $c := 1 div $c
   return
     fold-left(
       1 to $n, 0,
       function($z,$a) {
         $z,
         $z[last()] + $c div math:pow(xs:double($a), $alpha)
       }
     )[position()=(2 to last())] (: Skip starting 0 :)
};

(:
 : bsearch()
 : Service function to perform a binary search on an ordered sequence of values
 : Returns index of slot containing value 
 : 
 : $low: lower bound index
 : $mid: midpoint index
 : $high: upper bound index
 : $values: values to search
 : $key: value to find
 :)
declare %private function this:bsearch($low as xs:integer, $mid as xs:integer, $high as xs:integer, $values as xs:double*, $key as xs:double) as xs:integer
{
  if ($low gt $high) then $mid
  else (
    let $mid := ($low + $high) idiv 2
    return (
      if ($values[$mid] ge $key and $values[$mid - 1] lt $key) then $mid
      else if ($values[$mid] ge $key) then this:bsearch($low, $mid, $mid - 1, $values, $key)
      else this:bsearch($mid + 1, $mid, $high, $values, $key)
    )
  )
};

(:
 : bsearch()
 : Perform a binary search on an ordered sequence of values
 : Returns index of slot containing value 
 : Used by select-index() and markov()
 : 
 : $values: values to search
 : $key: value to find
 :)
declare function this:bsearch($values as xs:double*, $key as xs:double) as xs:integer
{
  let $n := count($values)
  return this:bsearch(1, (1 + $n) idiv 2, $n, $values, $key)
};

(:
 : select-index()
 : Return a rank from 1 to N of values distributed according to cumulative
 : probability sums. Use zipf-sums(), percent-sums(), etc. to compute those.
 :
 : $sums: cumulative probability sums, computed via zipf-sums() or percent-sums()
 : $n: maximum rank to return ($n should be <= count($sums))
 :) 
declare function this:select-index($sums as xs:double*, $n as xs:integer) as xs:integer
{
  (: Get value in (0,1) exclusive; underlying randomizer returns inclusive
   : values, so generate value within epsilon of bounds instead
   :)
  let $p := this:uniform($this:ε, 1 - $this:ε)
  return this:bsearch($sums[position()=(1 to $n)], $p)
};

(: 
 : selection()
 : Return a selection using data distributed per the cumulative probability sums.
 : 
 : $sums: cumulative probability sums, computed via zipf-sums(), percent-sums(), etc.
 : $values: values to select, in rank order
 :)
declare function this:selection($sums as xs:double*, $values as item()*) as item()
{
  let $n := min((count($sums),count($values)))
  let $k := this:select-index($sums, $n)
  return $values[$k]
};

(: 
 : check-markov-matrix()
 : Check whether the matrix is a valid Markov matrix, raise error if it is not.
 :)
declare function this:check-markov-matrix($dim as xs:integer, $matrix as xs:double*, $is-cumulative as xs:boolean) as empty-sequence()
{
  let $size := count($matrix)
  let $sqrt := math:sqrt($size)
  return (
    (: Matrix must be square :)
    if ($sqrt * $sqrt ne $size)
    then errors:error("RAND-NOTSQUARE", $size)
    else if ($dim ne ($sqrt cast as xs:integer))
    then errors:error("RAND-BADDIM", ($dim, $sqrt))
    else ()
    ,
    (: Values must be probabilities :)
    for $i in 1 to $size return (
      if ($matrix[$i] < 0 or $matrix[$i] > 1)
      then errors:error("RAND-BADPROBABILITY", ($i, $matrix[$i]))
      else ()
    )
    ,
    (: Each row should have a cumulative probability of 1 :)
    (: If $is-cumulative, each row has been expressed as cumulative 
     : probabilities already (for efficiency) so we check p<1 vs sum(p)<1
     : but we did that check above already.
     :)
    if ($is-cumulative) then () else (
      for $i in 1 to $dim return (
        let $sprob := sum($matrix[position()=(($i - 1)*$dim + 1 to ($i - 1) * $dim)])
        return (
          if (this:twixt($sprob, 0, 1 + $this:ε)) then ()
          else errors:error("RAND-BADSUMPROB", ($i, $sprob))
        )
      )
    )
  )
};

(:
 : markov-sums()
 : Return a Markov probability matrix as a matrix where each row contains
 : the cumulative probability sums for that row. This allows the Markov
 : distribution selection to run more efficiently.
 :
 : Example:
 :   markov-sums(3, (
 :     0.2, 0.4, 0.4,
 :     0.5, 0.5, 0,
 :     0.4, 0.2, 0.4
 :   )) => (
 :     0.2, 0.6, 1,
 :     0.5, 1, 1,
 :     0.4, 0.6, 1
 :   )
 :
 :  To account for rounding, the last item in the row is always forced to 1
 :  and any sums greater than 1 are also rounded down to 1.
 :
 : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
 : $matrix: the input non-cumulative Markov matrix
 :)
declare function this:markov-sums($dim as xs:integer, $matrix as xs:double*) as xs:double*
{
  for $i in 1 to $dim
  let $row := $matrix[position()=(($i - 1) * $dim + 1 to $i * $dim)]
  let $new-row :=
    fold-left(
      $row, 0,
      function($z,$a) {
        $z,
        min(($z[last()] + $a, 1))
      }
    )[position()=(2 to last())]
  return (
    if ($new-row[last()] lt 1) then ($new-row[position()=(1 to last()-1)],1) else $new-row
  )
};

(:
 : markov-percent-sums()
 : Return a Markov probability matrix as a matrix where each row contains
 : the cumulative probability sums for that row. This allows the Markov
 : distribution selection to run more efficiently.
 :
 : Example:
 :   markov-percent-sums(3, (
 :     20, 40, 40,
 :     50, 50, 0,
 :     40, 20, 40
 :   )) => (
 :     0.2, 0.6, 1,
 :     0.5, 1, 1,
 :     0.4, 0.6, 1
 :   )
 :
 :  To account for rounding, the last item in the row is always forced to 1
 :  and any sums greater than 1 are also rounded down to 1.
 :
 : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
 : $matrix: the input non-cumulative Markov matrix
 :)
declare function this:markov-percent-sums($dim as xs:integer, $matrix as xs:integer*) as xs:double*
{
  for $i in 1 to $dim
  let $row := $matrix[position()=(($i - 1) * $dim + 1 to $i * $dim)]
  let $new-row := 
    fold-left(
      $row, 0,
      function($z,$a) {
        $z,
        min(($z[last()] + ($a div 100), 1))
      }
    )[position()=(2 to last())]
  return (
    if ($new-row[last()] lt 1) then ($new-row[position()=(1 to last()-1)],1) else $new-row
  )
};

(:
 : markov()
 : Given a square matrix of Markov probabilities and a current state,
 : return a successor state with the probabilities as given. For efficiency,
 : the matrix should be given in the cumulative probability form, as created
 : by markov-sums() or markov-percent-sums()
 :
 : $current: Index of current state in the matrix
 : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
 : $matrix: Square matrix of cumulative probabilities
 :)
declare function this:markov($current as xs:integer, $dim as xs:integer, $matrix as xs:double*) as xs:integer
{
  let $_check :=
    if (this:twixt($current, 1, $dim)) then ()
    else errors:error("ML-BADARGS", ("current", $current))
  let $row := $matrix[position()=(($current - 1) * $dim + 1 to $current * $dim)]
  return this:select-index($row, $dim)
};

(: 
 : markov-selection()
 : Return a selection using Markov matrix distribution.
 : 
 : $current: index of current selection
 : $dim: size of each dimension of matrix, math:sqrt(count($matrix))
 : $matrix: cumulative probability matrix, computed via markov-sums()
 : $values: values to select, in index order
 :)
declare function this:markov-selection($current as xs:integer, $dim as xs:integer, $matrix as xs:double*, $values as item()*) as item()
{
  let $k := this:markov($current, $dim, $matrix)
  let $_check :=
    (: Make sure matrix and values are compatible :)
    if (this:twixt($k, 1, count($values))) then ()
    else errors:error("ML-BADARGS", ("values", $values))
  return $values[$k]
};


(: 
 : random-selection()
 : Return a random selection from a set of keys, where the key is selected
 : per a set of probabilities.
 : 
 : $weights: Map of probabilities (as integer percents) from key to weight
 : $default: Value to return if all else fails
 :)
declare function this:random-selection($weights as map(xs:string,item()*), $default as xs:string) as xs:string
{
  (
    for $key in this:shuffle($weights=>map:keys())
    where this:flip($weights=>map:get($key))
    return $key,
    $default
  )[1]
};

(:
 : random-assortment()
 : Select a subset of a set of keys, in (unform) random order, where each key
 : is selected at most once, and is selected with probability given by set
 : of weights.
 :
 : $keys: Set of keys to choose from
 : $weights: Map of probabilities (as integer percents) from key to weight
 :)
declare function this:random-assortment($keys as xs:string*, $weights as map(xs:string,item()*)) as xs:string*
{
  for $key in $keys
  return
    this:shuffle(
      if (this:flip($weights=>map:get($key)))
      then $key
      else ()
    )
};

(: 
 : cast()
 : Cast/format a value to a particular type
 : Service function for randomize()
 : 
 : Casting to decimal is actually just rounding to $DECIMAL-DIGITS fractional
 : digits.
 : 
 : $value: Value to cast
 : $cast-as: One of "integer", "decimal", "boolean", or "string"
 :)
declare function this:cast($value as item(), $cast-as as xs:string?)
{
  switch ($cast-as)
  case "integer" return $value cast as xs:integer
  case "decimal" return round($value * $this:DECIMAL-DIV) div $this:DECIMAL-DIV
  case "boolean" return boolean($value)
  case "string" return string($value)
  default return $value
};

(: 
 : range()
 : Ensure the value is in the range by changing value to minimum or maximum as
 : necessary. 
 : Service function for randomize()
 : Assumes min < max
 : 
 : $value: The value
 : $min: The lowest value (optional)
 : $max: The highest value (optional)
 :)
declare function this:range($value as xs:numeric, $min as xs:numeric?, $max as xs:numeric?) as xs:numeric
{
  let $low := if (exists($min)) then max(($min,$value)) else $value
  return
    if (exists($max)) then min(($max,$low)) else $low
};

(:
 : get-last()
 : Function available for callback use in fetch()
 : 
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   (see randomize())
 : $last: Previous value
 :)
declare function this:get-last($algorithm as map(xs:string,item()*), $last as item()?)
{
  $last
};

(:
 : fetch()
 : Lookup the key in the map. If it is a two-arg function or a QName of
 : a two-arg function, call that function, passing the algorithm map and
 : the last value to it. Otherwise return the value for the key.
 :
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   (see randomize())
 : $key: Key to lookup in map
 : $last: Previous value
 :)
declare %private function this:fetch($algorithm as map(xs:string,item()*), $key as xs:string, $last as item()?)
{
  let $v := $algorithm=>map:get($key)
  return
    typeswitch ($v)
    case function(map(*), item()?) as item()* return $v($algorithm, $last)
    case xs:QName return (
      let $f := function-lookup($v, 2)
      return $f($algorithm, $last)
    )
    default return $v
};

(:
 : is-complex()
 : Helper function for randomize() to allow for optimization of case where
 : we're using a static algorithm with no use of a function
 :)
declare %private function this:is-complex($algorithm as map(xs:string,item()*)) as xs:boolean
{
  if ($algorithm=>map:contains("is_complex")) then (
    $algorithm=>map:get("is_complex")
  ) else (
    some $k in map:keys($algorithm)
    satisfies (
      typeswitch (map:get($algorithm, $k))
      case xs:QName return true()
      case function(map(*), item()?) as item()* return true()
      default return false()
    )
  )
};

(:
 : is-complex()
 : Helper function for randomize() to allow for optimization of case where
 : we're using a static algorithm with no use of a function
 :)
declare %private function this:is-complex($key as xs:string, $algorithm as map(xs:string,item()*)) as xs:boolean
{
  typeswitch (map:get($algorithm, $key))
  case xs:QName return true()
  case function(map(*), item()?) as item()* return true()
  default return false()
};

(: 
 : modify-all()
 : Apply modifiers to all values. Assumes all key values are static.
 : Service routine for randomize()
 :)
declare %private function this:modify-all($values as item()*, $last as item()?, $algorithm as map(xs:string,item()*))
{
  let $distribution := $algorithm=>map:get("distribution")
  let $min := $algorithm=>map:get("min")
  let $max := $algorithm=>map:get("max")
  let $pre-multiplier := $algorithm=>map:get("pre-multiplier")
  let $values :=
    if (exists($pre-multiplier))
    then $values!(. * $pre-multiplier)    
    else $values
  let $values :=
    for $value in $values return (
      if ($distribution = ("uniform","flip","constant") or
        this:twixt($value, $min, $max)
      ) then (
        $value
      ) else (
        switch($algorithm=>map:get("truncation"))
        case "ceiling" return this:range($value, $min, $max)
        case "drop" return ()
        default (: "resample" :) return (
          let $new :=
            fold-left(
              1 to $this:RESAMPLE-LIMIT, (),
              function($v, $i) {
                if (empty($v)) then (
                  let $v := this:raw-distribution(1, $last, $algorithm)
                  let $v2 :=
                    if (exists($pre-multiplier))
                    then $v * $pre-multiplier
                    else $v
                  return
                    if (this:twixt($v2, $min, $max)) then $v
                    else ()
                 ) else (
                  $v
                 )
              }
            )
          return (
            if (empty($new)) then (
              trace((), "gave up resampling"),
              this:range($value, $min, $max)
            ) else (
              $new
            )
          )
        )
      )
    )
  let $values :=
    let $post-multiplier := $algorithm=>map:get("post-multiplier")
    return
      if (exists($post-multiplier))
      then $values!(. * $post-multiplier)
      else $values
  let $values :=
    let $cast-as := $algorithm=>map:get("cast")
    return
      if (exists($cast-as))
      then $values!this:cast(., $cast-as)
      else $values
  return $values
};

(:
 : modify()
 : Apply modifiers to values. Assumes all key values may be dynamic.
 : Service routine for randomize()
 :) 
declare %private function this:modify($value as item(), $last as item()?, $algorithm as map(xs:string,item()*))
{
  let $distribution := $algorithm=>this:fetch("distribution",$last)
  let $min := $algorithm=>this:fetch("min",$last)
  let $max := $algorithm=>this:fetch("max",$last)
  let $pre-multiplier := $algorithm=>this:fetch("pre-multiplier", $last)
  let $value :=
    if (exists($pre-multiplier))
    then $value * $pre-multiplier
    else $value
  let $value :=
    if ($distribution = ("uniform","flip","constant") or
      this:twixt($value, $min, $max)
    ) then (
      $value
    ) else (
      switch($algorithm=>this:fetch("truncation",$last))
      case "drop" return ()
      case "ceiling" return this:range($value, $min, $max)
      default (: "resample" :) return (
        let $new :=
          fold-left(
            1 to $this:RESAMPLE-LIMIT, (),
            function($v, $i) {
              if (empty($v)) then (
                let $v := this:raw-distribution(1, $last, $algorithm)
                let $v2 :=
                  if (exists($pre-multiplier))
                  then $v * $pre-multiplier
                  else $v
                return
                  if (this:twixt($v2, $min, $max)) then $v
                  else ()
               ) else (
                $v
               )
            }
          )
        return (
          if (empty($new)) then (
            trace((),"gave up resampling"),
            this:range($value, $min, $max)
          ) else (
            $new
          )
        )
      )
    )
  let $value :=
    let $post-multiplier := $algorithm=>this:fetch("post-multiplier",$last)
    return
      if (exists($post-multiplier))
      then $value * $post-multiplier
      else $value
  let $value :=
    let $cast-as := $algorithm=>this:fetch("cast",$last)
    return
      if (exists($cast-as))
      then this:cast($value, $cast-as)
      else $value
  return $value
};

(:
 : raw-distribution()
 : Service function to randomize(). Creates the raw values from the 
 : distributions, without applying the modifications (min/max, multipliers,
 : casts)
 :
 : Assumes distribution is fixed (or distributions for multi-modal)
 : $N: How many values to return
 : $last: Previous value (needed for markov, available to key callback)
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   distribution: Distribution to use, one of "constant", "uniform", "normal",
 :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal"
 :   min: mimumum value (uniform, constant)
 :   max: maximum value (uniform, constant)
 :   mean: mean of distribution (normal, skewed) (default=0)
 :   std: standard deviation of distribution (normal, skewed) (default=1)
 :   skew: skew of distribution (skewed) (default=0)
 :   p: probability (bernoulli, flip) (default=50)
 :   sums: cumulative probability sums (zipf, markov, sums)
 :   alpha: alpha parameter (zipf) (needed if no sums) (default=1)
 :   limit: number of sums (zipf) (needed if no sums) (default=1000)
 :   start: index of starting symbol (markov) (default=uniform[1,dim])
 :   dim: size of each dimension of Markov matrix (markov)
 :   matrix: raw Markov matrix (used if sums not provided, not recommended)
 :   distributions: sequence of distributions to combine (multimodal)
 :)
declare %private function this:raw-distribution($N as xs:integer, $last as item()?, $algorithm as map(xs:string,item()*))
{
  if (this:is-complex($algorithm)) then (
    (: Distribution is fixed, but we may alter any other parameter 
     : so need to feed $last as appropriate to the sequence
     : For multimodal distributions is also fixed
     :)
    let $distribution := ($algorithm=>map:get("distribution"),"uniform")[1]
    let $values := 
      switch ($distribution)
      case "constant" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $min := $algorithm=>this:fetch("min",$values[last()])
              let $max := $algorithm=>this:fetch("max",$values[last()])
              let $val := this:constant($min, $max)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "uniform" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $min := $algorithm=>this:fetch("min",$values[last()])
              let $max := $algorithm=>this:fetch("max",$values[last()])
              let $val := this:uniform($min, $max)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "normal" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $μ := ($algorithm=>this:fetch("mean",$last),0)[1]
              let $σ := ($algorithm=>this:fetch("std",$last),1)[1]
              let $val := this:normal($μ, $σ)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "skewed" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $μ := ($algorithm=>this:fetch("mean",$last),0)[1]
              let $σ := ($algorithm=>this:fetch("std",$last),1)[1]
              let $α := ($algorithm=>this:fetch("skew",$last),0)[1]
              let $val := this:skewed($μ, $σ, $α)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "bernoulli" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $p := ($algorithm=>this:fetch("p",$last),50)[1]
              let $val := if (this:flip($p)) then 1 else 0  
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "flip" return
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $last := $values[last()]
              let $p := ($algorithm=>this:fetch("p",$last),50)[1]
              let $val := this:flip($p)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "multimodal" return (
        (: multimodal has sequence of sub randomizers in 'distributions' :)
        (: combined in equal fractions :)
        (: $last applies to each subrandomizer independently generally
         : except post-sum modifications apply to combined last
         :)
        let $distributions := $algorithm=>map:get("distributions")
        let $numd := count($distributions)
        let $flips := for $i in 1 to $N return this:cast(this:uniform(1, $numd),"integer")
        (: RAW: v11, v12, ..., v1N, v21, ..., v2N, ...., vd1, ... vdN :)
        let $raw :=
          for $subdistribution in $distributions
          return this:randomize($N, $last, $subdistribution)
        let $result :=           
          for $i in 1 to $N
          return $raw[($flips[$i] - 1)*$N + $i]
        return
          $result
      )
      case "sums" return
        (: It's probably a bad idea to have non-static probability sums :)
        (: Precompute these and include them in the algorithm map :)
        let $fixed-sums :=
          if (this:is-complex("sums", $algorithm)) then ()
          else $algorithm=>map:get("sums")
        let $fixed-max-rank :=
          if (exists($fixed-sums)) then count($fixed-sums) else ()
        let $result := 
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $sums :=
                if (exists($fixed-sums)) then $fixed-sums
                else $algorithm=>this:fetch("sums",$values[last()])
              let $max-rank :=
                if (exists($fixed-max-rank))
                then $fixed-max-rank
                else count($sums)
              let $val := this:select-index($sums, $max-rank)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "zipf" return
        (: It's probably a bad idea to have non-static Zipf parameters
         : An even worse idea to have to recompute probability sums 
         : for each time through.
         : Precompute these and put them in the map or you'll be sorry
         :)
        let $fixed-sums :=
          if (this:is-complex("sums", $algorithm)) then ()
          else (
            if (exists($algorithm=>map:get("sums")))
            then $algorithm=>map:get("sums")
            else (
              if (this:is-complex("alpha", $algorithm) or
                  this:is-complex("limit", $algorithm)
              ) then ()
              else (
                this:zipf-sums(
                  ($algorithm=>map:get("alpha"),1)[1],
                  ($algorithm=>map:get("limit"),1000)[1]
                )
              )
            )
          )
        let $fixed-max-rank :=
          if (exists($fixed-sums)) then count($fixed-sums) else ()
        let $result := 
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $sums :=
                if (exists($fixed-sums)) then $fixed-sums
                else if (exists($algorithm=>this:fetch("sums",$values[last()])))
                then $algorithm=>this:fetch("sums",$values[last()])
                else
                  this:zipf-sums(
                    ($algorithm=>this:fetch("alpha",$values[last()]),1)[1],
                    ($algorithm=>this:fetch("limit",$values[last()]),1000)[1]
                  )
              let $max-rank :=
                if (exists($fixed-max-rank))
                then $fixed-max-rank
                else count($sums)
              let $val := this:select-index($sums, $max-rank)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      case "markov" return
        (: It's probably a bad idea to have non-static Markov parameters :)
        (: Precompute them and include them in the algorithm map :)
        let $fixed-sums :=
          if (this:is-complex("sums", $algorithm)) then ()
          else (
            if (exists($algorithm=>map:get("sums")))
            then $algorithm=>map:get("sums")
            else (
              if (this:is-complex("matrix", $algorithm) or
                  this:is-complex("dim", $algorithm)
              ) then ()
              else (
                let $matrix := $algorithm=>map:get("matrix")
                let $dim := (
                  $algorithm=>map:get("dim"),
                  math:sqrt(count($matrix)) cast as xs:integer
                )[1]
                return this:markov-sums($dim, $matrix)
              )
            )
          )
        let $fixed-dim :=
          if (this:is-complex("dim", $algorithm)) then ()
          else if ($algorithm=>map:contains("dim"))
          then $algorithm=>map:get("dim")
          else if (exists($fixed-sums))
          then math:sqrt(count($fixed-sums)) cast as xs:integer
          else ()
        let $start := ($last,$algorithm=>this:fetch("start",$last),this:cast(this:uniform(1,$fixed-dim),"integer"))[1]
        return
          fold-left(
            1 to $N, $start,
            function($values,$a) {
              $values
              ,
              let $sums :=
                if (exists($fixed-sums)) then $fixed-sums
                else if (exists($algorithm=>this:fetch("sums",$values[last()])))
                then $algorithm=>this:fetch("sums",$values[last()])
                else (
                  let $matrix := $algorithm=>this:fetch("matrix",$values[last()])
                  let $dim := (
                    $algorithm=>this:fetch("dim",$values[last()]),
                    math:sqrt(count($matrix)) cast as xs:integer
                  )[1]
                  return this:markov-sums($dim, $matrix)
                )
              let $dim := (
                $fixed-dim,
                $algorithm=>this:fetch("dim",$values[last()]),
                math:sqrt(count($sums)) cast as xs:integer
              )[1]
              let $val := this:markov($values[last()],$dim,$sums)
              return $val
            }
          )[position()=(2 to last())] (: start is never empty :)
      default return (
        let $result :=
          fold-left(
            1 to $N, $last,
            function($values, $i) {
              $values
              ,
              let $min := $algorithm=>this:fetch("min",$values[last()])
              let $max := $algorithm=>this:fetch("max",$values[last()])
              let $val := this:constant($min, $max)
              return $val
            }
          )
        return
          if (empty($last)) then $result else $result[position()=(2 to last())]
      )
    return $values
  ) else (
    (: Everything is fixed, just look up keys :)
    let $distribution := ($algorithm=>map:get("distribution"),"uniform")[1]
    let $min := $algorithm=>map:get("min") (: OK to be empty :)
    let $max := $algorithm=>map:get("max") (: OK to be empty :)
    let $values := 
      switch ($distribution)
      case "constant" return
        for $i in 1 to $N
        return this:constant($min, $max)
      case "uniform" return
        for $i in 1 to $N
        return this:uniform($min, $max)
      case "normal" return
        let $μ := ($algorithm=>map:get("mean"),0)[1]
        let $σ := ($algorithm=>map:get("std"),1)[1]
        for $i in 1 to $N
        return this:normal($μ, $σ)
      case "skewed" return
        let $μ := ($algorithm=>map:get("mean"),0)[1]
        let $σ := ($algorithm=>map:get("std"),1)[1]
        let $α := ($algorithm=>map:get("skew"),0)[1]
        for $i in 1 to $N
        return this:skewed($μ, $σ, $α)
      case "bernoulli" return
        let $p := ($algorithm=>map:get("p"),50)[1]
        for $i in 1 to $N
        return 
          if (this:flip($p)) then 1 else 0
      case "flip" return
        let $p := ($algorithm=>map:get("p"),50)[1]
        for $i in 1 to $N
        return this:flip($p)
      case "sums" return
        let $sums := $algorithm=>map:get("sums")
        let $max-rank := count($sums)
        for $i in 1 to $N
        return this:select-index($sums, $max-rank)
      case "multimodal" return (
        let $distributions := $algorithm=>map:get("distributions")
        let $numd := count($distributions)
        let $flips := for $i in 1 to $N return this:cast(this:uniform(1, $numd),"integer")
        (: RAW: v11, v12, ..., v1N, v21, ..., v2N, ...., vd1, ... vdN :)
        let $raw :=
          for $subdistribution in $distributions
          return this:randomize($N, $last, $subdistribution)
        for $i in 1 to $N
        return (
          $raw[($flips[$i] - 1)*$N + $i]
        )
      )
      case "zipf" return
        let $sums :=
          if (exists($algorithm=>map:get("sums")))
          then $algorithm=>map:get("sums")
          else
            this:zipf-sums(
              ($algorithm=>map:get("alpha"),1)[1],
              ($algorithm=>map:get("limit"),1000)[1]
            )
        let $max-rank := count($sums)
        for $i in 1 to $N
        return this:select-index($sums, $max-rank)
      case "markov" return
        let $sums :=
          if (exists($algorithm=>map:get("sums")))
          then $algorithm=>map:get("sums")
          else (
            let $matrix := $algorithm=>map:get("matrix")
            let $dim := (
              $algorithm=>map:get("dim"),
              math:sqrt(count($matrix)) cast as xs:integer
            )[1]
            return this:markov-sums($dim, $matrix)
          )
        let $dim := (
          $algorithm=>map:get("dim"),
          math:sqrt(count($sums)) cast as xs:integer
        )[1]
        let $start := ($last,$algorithm=>map:get("start"),this:cast(this:uniform(1,$dim),"integer"))[1]
        return
          fold-left(
            1 to $N, $start,
            function($z,$a) {
              $z,
              this:markov($z[last()],$dim,$sums)
            }
          )[position()=(2 to last())]
      default return
        for $i in 1 to $N return this:constant($min, $max)
    return $values
  )
};

(: 
 : select-random()
 : Given a set of values and a randomizer (see randomize()), use the output
 : of randomize as an index into the value set, and return the selected values.
 : 
 : $N: How many values to return
 : $last: Previous value (needed for markov, available to key callback)
 : $algorithm: The algorithm to use, a map with various parameter keys
 : $values: The value set to choose from
 :)
declare function this:select-random($N as xs:integer, $last as item()?, $algorithm as map(xs:string,item()*), $values as item()*) as item()*
{
  let $count := count($values)
  let $indices := this:randomize($N, $last, $algorithm)
  for $raw-index in $indices
  (: let $index := 1 + (($raw-index cast as xs:integer) mod $count) :)
  let $index := max((min(($raw-index cast as xs:integer, $count)),1))
  return $values[$index]
};

(: 
 : select-random()
 : Given a set of values and a randomizer (see randomize()), use the output
 : of randomize as an index into the value set, and return the selected values.
 : 
 : $N: How many values to return
 : $algorithm: The algorithm to use, a map with various parameter keys
 : $values: The value set to choose from
 :)
declare function this:select-random($N as xs:integer, $algorithm as map(xs:string,item()*), $values as item()*) as item()*
{
  let $count := count($values)
  let $indices := this:randomize($N, (), $algorithm)
  for $raw-index in $indices
  let $index := max((min(($raw-index cast as xs:integer, $count)),1))
  return $values[$index]
};

(: 
 : select-random()
 : Given a set of values and a randomizer (see randomize()), use the output
 : of randomize as an index into the value set, and return the selected values
 : 
 : $algorithm: The algorithm to use, a map with various parameter keys
 : $values: The value set to choose from
 :)
declare function this:select-random($algorithm as map(xs:string,item()*), $values as item()*) as item()
{
  let $count := count($values)
  let $raw-index := this:randomize($algorithm)
  let $index := max((min(($raw-index cast as xs:integer, $count)),1))
  return $values[$index]
};

(: 
 : select-random()
 : Given a set of values pick as (uniformly random) value from the
 : value set, and return it
 : 
 : $algorithm: The algorithm to use, a map with various parameter keys
 : $values: The value set to choose from
 :)
declare function this:select-random($values as item()*) as item()
{
  let $count := count($values)
  let $algorithm :=
    dist:uniform(1, $count)=>dist:cast("integer")=>dist:is-complex(false())
  let $raw-index := this:randomize($algorithm)
  (: let $index := 1 + (($raw-index cast as xs:integer) mod $count) :)
  let $index := max((min(($raw-index cast as xs:integer, $count)),1))
  return $values[$index]
};

(: 
 : randomize()
 : Return a sequence of random data distributed as per the algorithm.
 : 
 : $N: How many values to return
 : $last: Previous value (needed for markov, available to key callback)
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   distribution: Distribution to use, one of "constant", "uniform", "normal",
 :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal"
 :   min: minimum value (optional)
 :   max: maximum value (optional)
 :   pre-multiplier: multiplier before min/max (optional)
 :   post-multiplier: multiplier after min/max (optional)
 :   cast: cast type (optional)
 :   mean: mean of distribution (normal, skewed) (default=0)
 :   std: standard deviation of distribution (normal, skewed) (default=1)
 :   skew: skew of distribution (skewed) (default=0)
 :   p: probability (bernoulli, flip) (default=50)
 :   sums: cumulative probability sums (zipf, markov, sums)
 :   alpha: alpha parameter (zipf) (needed if no sums) (default=1)
 :   limit: number of sums (zipf) (needed if no sums) (default=1000)
 :   start: index of starting symbol (markov) (default=uniform[1,dim])
 :   dim: size of each dimension of Markov matrix (markov)
 :   matrix: raw Markov matrix (used if sums not provided, not recommended)
 :   distributions: sequence of distributions to combine (multimodal)
 :   truncation: how to keep value in [min,max] range (not constant/uniform)
 :     ceiling => ceiling/floor to min/max
 :     drop => toss out (so may return < $N)
 :     resample (default) => resample until an in-bound value is found or
 :       resample limit is reached ($RESAMPLE-LIMIT=10) then use ceiling
 :)
declare function this:randomize($N as xs:integer, $last as item()?, $algorithm as map(xs:string,item()*))
{
  if (this:is-complex($algorithm)) then (
    if (this:is-complex("distribution", $algorithm) or
        this:is-complex("distributions", $algorithm)
    ) then (
      (: We change distributions, so we have to do this one value at a time :)
      let $result :=
        fold-left(
          1 to $N, $last,
          function($vals, $i) {
            $vals
            ,
            this:randomize(1, $vals[last()], $algorithm)
          }
        )
      return (
        if (empty($last)) then $result
        else $result[position()=(2 to last())]
      )
    ) else (
      (: Distribution is fixed, but we may alter any other parameter 
       : so need to feed $last as appropriate to the sequence
       : For multimodal distributions is also fixed
       :)
      let $distribution := ($algorithm=>map:get("distribution"),"uniform")[1]
      let $values := this:raw-distribution($N, $last, $algorithm)
      let $values :=
        fold-left(
          $values, $last,
          function($vals, $val) {
            $vals
            ,
            this:modify($val, $vals[last()], $algorithm)
          }
        )
      return
        if (empty($last)) then $values
        else $values[position()=(2 to last())]
    )
  ) else (
    (: Everything is fixed, just look up keys :)
    let $distribution := ($algorithm=>map:get("distribution"),"uniform")[1]
    let $values := this:raw-distribution($N, $last, $algorithm)
    let $values := this:modify-all($values, $last, $algorithm)
    return (
      $values
    )
  )
};

(: 
 : randomize()
 : Generate one random value with given previous value.
 : See randomize($N, $last, $algorithm) above.
 :)
declare function this:randomize($N as xs:integer, $algorithm as map(xs:string,item()*))
{
  this:randomize($N, (), $algorithm)
};

(: 
 : randomize()
 : Generate one random value.
 : See randomize($N, $last, $algorithm) above.
 :)
declare function this:randomize($algorithm as map(xs:string,item()*))
{
  this:randomize(1, (), $algorithm)
};

(:
 : twixt()
 : Return true of the value if between the minimum and maximum.
 :
 : $value: Value to test
 : $min: Minimum (inclusive)
 : $max: Maximum (inclusive)
 :)
declare function this:twixt($value as xs:double, $min as xs:double?, $max as xs:double?)
{
  (if (exists($min)) then ($value ge $min) else true()) and
  (if (exists($max)) then ($value le $max) else true())
};


(: 
 : histogram()
 : Print out the distribution as a histogram.
 : 
 : $width: Size of each bucket (max - min)
 : $height: Number of values per star
 : $data: Data to analyze
 :)
declare function this:histogram($width as xs:integer, $height as xs:integer, $data as xs:double*) as xs:string*
{
  let $buckets :=
    fold-left(
      $data, map {},
      function($buckets as map(xs:string, xs:integer), $d as xs:double) {
        let $key := format-number($d div $width,"0")
        let $key := if ($key="-0") then "0" else $key (: avoid 2 buckets on 0 :)
        let $num := $buckets=>map:get($key)
        return (
          if (exists($num))
          then $buckets=>map:put($key, xs:integer($num)+1)
          else $buckets=>map:put($key, 1)
        )
      }
    )
  for $key in $buckets=>map:keys()
  let $num := xs:integer($buckets=>map:get($key))
  order by xs:integer($key)
  return (
    format-number(xs:integer($key)*$width," 00;-00")||" "||
    format-number($num,"000")||" "||
    string-join(for $i in 1 to $num idiv $height return "*","")
  )
};

(: 
 : mean()
 : Return the mean of the distribution. Some distributions don't have a mean;
 : these will raise an error.
 : 
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   (see randomize())
 : $last: Previous value
 :)
declare function this:mean($algorithm as map(xs:string,item()*), $last as item()?)
{
  switch ($algorithm=>map:get("distribution"))
  case "constant" return this:randomize($algorithm)
  case "uniform" return
    let $min := $algorithm=>this:fetch("min",$last) (: OK to be empty :)
    let $max := $algorithm=>this:fetch("max",$last) (: OK to be empty :)
    return
      if ($max < $min) then errors:error("RAND-BADMINMAX",($min,$max))
      else ($max - $min) idiv 2
  case "normal" return ($algorithm=>this:fetch("mean",$last),0)[1]
  case "skewed" return ($algorithm=>this:fetch("mean",$last),0)[1]
  case "bernoulli" return
    let $p := ($algorithm=>this:fetch("p",$last),50)[1]
    return $p div 100
  case "flip" return
    let $p := ($algorithm=>this:fetch("p",$last),50)[1]
    return $p div 100
  default return errors:error("RAND-UNIMPLEMENTED", ($algorithm=>map:get("distribution"), "mean"))
};

(: 
 : mean()
 : Return the mean of the distribution. Some distributions don't have a mean;
 : these will raise an error.
 : 
 : $algorithm: The algorithm to use, a map with various parameter keys
 :   (see randomize())
 :)
declare function this:mean($algorithm as map(xs:string,item()*))
{
  this:mean($algorithm, ())
};

(: 
 : uniform-distribution-of()
 : Construct an algorithm map for a uniform distribution of index keys
 : (1 to N) over the given set of values.
 : 
 : $items: the set of values to contruct a uniform index distribution over
 :)
declare function this:uniform-distribution-of($items as item()*) as map(xs:string,item()*)
{
  dist:uniform(1, count($items))=>dist:cast("integer")
};

(: random.xqy :)
