xquery version "3.1";
(:
 : Module with functions providing sparse array operations.
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/core/sparse"; 

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

declare variable $this:RESERVED := ("zero");

(: 
 : array()
 : Constructor for sparse array.
 :
 : $zero: Zero value to use for missing entries.
 :)
declare function this:array($zero as item()*) as map(*)
{
  map {
    "zero": $zero
  }
};

(: 
 : array()
 : Constructor for sparse array, using 0 as the zero value for missing entries.
 :)
declare function this:array() as map(*)
{
  map {
    "zero": 0
  }
};

(:
 : put()
 : Put an entry at the nth slot in the sparse array, returning the new array
 :
 : $sparse: the sparse array
 : $n: the slot number
 : $value: the value to put in that slot
 :)
declare function this:put(
  $sparse as map(*),
  $n as xs:integer,
  $value as item()
) as map(*)
{
  $sparse=>map:put($n, $value)
};

(: 
 : get()
 : Get the value from the nth slot in the array
 :
 : $sparse: the sparse array
 : $n: the slot number
 :)
declare function this:get($sparse as map(*), $n as xs:integer) as item()
{
  if ($sparse=>map:contains($n))
  then $sparse($n)
  else $sparse("zero")
};

(:
 : delete()
 : Remove the data from the nth slot in the sparse array (i.e. set it to 
 : the zero value), returning the new array
 :
 : $sparse: the sparse array
 : $n: the slot number
 :)
declare function this:delete($sparse as map(*), $n as xs:integer) as map(*)
{
  $sparse=>map:remove($n)
};

(:
 : maybe-put-zero()
 : Insert a the zero value in the given slot if the value
 : is not present, returning the new array.
 :
 : $sparse: the sparse array
 : $n: the slot number
 :)
declare function this:maybe-put-zero(
  $sparse as map(*),
  $n as xs:integer
) as map(*)
{
  if ($sparse=>map:contains($n)) then $sparse
  else $sparse=>map:put($n, $sparse("zero"))
};

(:
 : is-set()
 : Does the nth slot have a value? 
 :
 : $sparse: the sparse array
 : $n: the slot number
 :)
declare function this:is-set($sparse as map(*), $n as xs:integer) as xs:boolean
{
  $sparse=>map:contains($n)
};

(:
 : collection()
 : Return all the non-zero values in the array as a sequence
 :
 : $sparse: the sparse array
 :)
declare function this:collection($sparse as map(*)) as item()*
{
  (($sparse=>map:keys())[not(string(.) = $this:RESERVED)])!($sparse(.))
};

(:
 : values()
 : Return all the values in the array between the given indices,
 : including zero values in the range.
 : 
 : $sparse: the sparse array
 : $start: lower bound
 : $end: upper bound
 :)
declare function this:values($sparse as map(*), $start as xs:integer, $end as xs:integer) as item()*
{
  let $zero := $sparse("zero")
  for $i in $start to $end
  return ($sparse($i),$zero)[1]
};

(:
 : values()
 : Return all the values in the array between the minimum and maximum
 : indices actually used, including zero values in the range.
 : 
 : $sparse: the sparse array
 :)
declare function this:values($sparse as map(*)) as item()*
{
  let $zero := $sparse("zero")
  let $keys := (($sparse=>map:keys())[not(string(.) = $this:RESERVED)])!number(.)
  for $i in (min($keys) cast as xs:integer) to (max($keys) cast as xs:integer)
  return ($sparse($i),$zero)[1]
};

(: 
 : min-index()
 : Return the bottom active index
 : 
 : $sparse: the sparse array
 :)
declare function this:min-index($sparse as map(*)) as xs:integer
{
  let $keys := (($sparse=>map:keys())[not(string(.) = $this:RESERVED)])!number(.)
  return (min($keys) cast as xs:integer)
};

(: 
 : max-index()
 : Return the top active index
 : 
 : $sparse: the sparse array
 :)
declare function this:max-index($sparse as map(*)) as xs:integer
{
  let $keys := (($sparse=>map:keys())[not(string(.) = $this:RESERVED)])!number(.)
  return (max($keys) cast as xs:integer)
};

(: sparse :)
