xquery version "3.1";
(:
 : Module with functions providing some basic utility operations.
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/core/utilities"; 

import module namespace errors="http://mathling.com/core/errors"
       at "errors.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";
declare namespace saxon="http://saxon.sf.net/";

(:======================================================================
 : Mathematical operations
 :======================================================================:)
 
(:
 : The first 100 primes
 :)
declare variable $this:PRIMES100 as xs:integer* :=
(
  2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509, 521, 523, 541
);

(:
 : decimal()
 : Cast the value to value with the given number of digits.
 : Useful for SVG compaction.
 :
 : $value: numeric value
 : $digits: how many digits to keep after the decimal point 
 :)
declare function this:decimal($value as xs:double, $digits as xs:integer) as xs:double
{
  switch ($digits)
  case 0 return round($value) cast as xs:double
  case 1 return (round($value * 10) div 10) cast as xs:double
  case 2 return (round($value * 100) div 100) cast as xs:double
  default return 
    let $div := math:pow(10,$digits) cast as xs:double
    return (round($value * $div) cast as xs:double div $div) cast as xs:double
};

(:
 : round()
 : Round and return an integer. Avoid problems when parameter wants an
 : integer and chokes on the double fn:round gives you.
 :
 : $value: the value to round
 :)
declare function this:round($value as xs:numeric) as xs:integer
{
  fn:round($value) cast as xs:integer
};

declare variable $this:UINT64_MAX as xs:unsignedLong := 18446744073709551615;

(:
 : unsignedLong()
 : Cast to xs:unsignedLong, but treat negative numbers as overflowed, not
 : as errors. For bit hackery manipulations e.g. in noise functions.
 :)
declare function this:unsignedLong($x as xs:integer) as xs:unsignedLong
{
  if ($x < 0) then $this:UINT64_MAX + $x + 1 else xs:unsignedLong($x)
};

declare variable $this:MULTIPLIERS64 as xs:integer* :=
  reverse(for $i in 0 to 64 return fn:round(math:pow(2, $i)) cast as xs:integer)
;

(:
 : Figure out how we can do XOR
 : Worst case: super expensive math 
 :)
declare %private variable $this:XOR-IMPL as function(xs:unsignedLong, xs:unsignedLong) as xs:unsignedLong :=
  let $xor64 := function-lookup(QName("http://marklogic.com/xdmp/", "xor64"), 2)
  let $pack := function-lookup(QName("http://expath.org/ns/binary","pack-integer"), 2)
  let $unpack := function-lookup(QName("http://expath.org/ns/binary","unpack-unsigned-integer"), 3)
  let $xor := function-lookup(QName("http://expath.org/ns/binary","xor"), 2)
  return (
    if (exists($xor64)) then (
      $xor64
    ) else if (exists($pack) and exists($xor) and exists($unpack)) then (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        $unpack(
          $xor(
            $pack($x,4),
            $pack($y,4)
          ),0,4) cast as xs:unsignedLong
      }
    ) else (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        let $xbits := for $i in $this:MULTIPLIERS64 return floor($x div $i) mod 2
        let $ybits := for $i in $this:MULTIPLIERS64 return floor($y div $i) mod 2
        let $xorbits :=
          for $bit at $i in $xbits return (
            xs:integer(($bit or $ybits[$i]) and not($bit=$ybits[$i]))
          )
        return (
          sum(
            for $bit at $i in $xorbits return $this:MULTIPLIERS64[$i]*$bit
          ) cast as xs:unsignedLong
        )
      }
    )
  )
;

(:
 : xor()
 : Bitwise XOR of two 64-bit unsigned long values.
 : You really don't want to be calling this much unless you have xdmp:xor64
 : or bin:xor.
 :)
declare function this:xor($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
{
  $this:XOR-IMPL($x, $y)
};

(:
 : Figure out how we can do bitwise OR
 : Worst case: super expensive math 
 :)
declare %private variable $this:OR-IMPL as function(xs:unsignedLong, xs:unsignedLong) as xs:unsignedLong :=
  let $or64 := function-lookup(QName("http://marklogic.com/xdmp/", "or64"), 2)
  let $pack := function-lookup(QName("http://expath.org/ns/binary","pack-integer"), 2)
  let $unpack := function-lookup(QName("http://expath.org/ns/binary","unpack-unsigned-integer"), 3)
  let $or := function-lookup(QName("http://expath.org/ns/binary","or"), 2)
  return (
    if (exists($or64)) then (
      $or64
    ) else if (exists($pack) and exists($or) and exists($unpack)) then (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        $unpack($or($pack($x,4),$pack($y,4)),0,4) cast as xs:unsignedLong
      }
    ) else (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        let $orbits :=
          for $i in $this:MULTIPLIERS64 return (
            xs:integer((floor($x div $i) mod 2) or (floor($y div $i) mod 2))
          )
        return (
          sum(
            for $bit at $i in $orbits return $this:MULTIPLIERS64[$i]*$bit
          ) cast as xs:unsignedLong
        )
      }
    )
  )
;

(:
 : or()
 : Bitwise OR of two 64-bit unsigned long values.
 : You really don't want to be calling this much unless you have xdmp:or64
 : or bin:or.
 :)
declare function this:or($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
{
  $this:OR-IMPL($x, $y)
};

(:
 : Figure out how we can do bitwise AND
 : Worst case: super expensive math 
 :)
declare %private variable $this:AND-IMPL as function(xs:unsignedLong, xs:unsignedLong) as xs:unsignedLong :=
  let $and64 := function-lookup(QName("http://marklogic.com/xdmp/", "and64"), 2)
  let $pack := function-lookup(QName("http://expath.org/ns/binary","pack-integer"), 2)
  let $unpack := function-lookup(QName("http://expath.org/ns/binary","unpack-unsigned-integer"), 3)
  let $and := function-lookup(QName("http://expath.org/ns/binary","and"), 2)
  return (
    if (exists($and64)) then (
      $and64
    ) else if (exists($pack) and exists($and) and exists($unpack)) then (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        $unpack($and($pack($x,4),$pack($y,4)),0,4) cast as xs:unsignedLong
      }
    ) else (
      function($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
      {
        let $andbits :=
          for $i in $this:MULTIPLIERS64 return (
            xs:integer((floor($x div $i) mod 2) and (floor($y div $i) mod 2))
          )
        return (
          sum(
            for $bit at $i in $andbits return $this:MULTIPLIERS64[$i]*$bit
          ) cast as xs:unsignedLong
        )
      }
    )
  )
;

(:
 : and()
 : Bitwise AND of two 64-bit unsigned long values.
 : You really don't want to be calling this much unless you have xdmp:and64
 : or bin:and.
 :)
declare function this:and($x as xs:unsignedLong, $y as xs:unsignedLong) as xs:unsignedLong
{
  $this:AND-IMPL($x, $y)
};

(:
 : Figure out how we can do LSHIFT
 :)
declare %private variable $this:LSHIFT-IMPL as function(xs:unsignedLong, xs:integer) as xs:unsignedLong :=
  let $lshift64 := function-lookup(QName("http://marklogic.com/xdmp/", "lshift64"), 2)
  let $bshift := function-lookup(QName("http://expath.org/ns/binary","shift"), 2)
  let $pack := function-lookup(QName("http://expath.org/ns/binary","pack-integer"), 2)
  let $unpack := function-lookup(QName("http://expath.org/ns/binary","unpack-unsigned-integer"), 3)
  return (
    if (exists($lshift64)) then (
      $lshift64
    ) else if (exists($pack) and exists($bshift) and exists($unpack)) then (
      function($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
      {
        $unpack($bshift($pack($x,4),$shift),0,4) cast as xs:unsignedLong
      }
    ) else (
      function($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
      {
        ($x * math:pow(2, $shift)) cast as xs:unsignedLong
      }
    )
  )
;

(:
 : lshift()
 : Bitwise left shift of 64-bit unsigned long value.
 :)
declare function this:lshift($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
{
  $this:LSHIFT-IMPL($x, $shift)
};

(:
 : Figure out how we can do RSHIFT
 :)
declare %private variable $this:RSHIFT-IMPL as function(xs:unsignedLong, xs:integer) as xs:unsignedLong :=
  let $rshift64 := function-lookup(QName("http://marklogic.com/xdmp/", "rshift64"), 2)
  let $bshift := function-lookup(QName("http://expath.org/ns/binary","shift"), 2)
  let $pack := function-lookup(QName("http://expath.org/ns/binary","pack-integer"), 2)
  let $unpack := function-lookup(QName("http://expath.org/ns/binary","unpack-unsigned-integer"), 3)
  return (
    if (exists($rshift64)) then (
      $rshift64
    ) else if (exists($pack) and exists($bshift) and exists($unpack)) then (
      function($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
      {
        $unpack($bshift($pack($x,4),-$shift),0,4) cast as xs:unsignedLong
      }
    ) else (
      function($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
      {
        floor($x * math:pow(2, -$shift)) cast as xs:unsignedLong
      }
    )
  )
;

(:
 : rshift()
 : Bitwise right shift of 64-bit unsigned long value.
 :)
declare function this:rshift($x as xs:unsignedLong, $shift as xs:integer) as xs:unsignedLong
{
  $this:RSHIFT-IMPL($x, $shift)
};

(:
 : modix()
 : Index into a sequence modulo sequence size. For the moral equivalent of the C
 : programmer's some_array[i % n_values]
 : 
 : $value: input index
 : $n: size of target sequence
 :)
declare function this:modix($value as xs:integer, $n as xs:integer) as xs:integer
{
  1 + ($value + $n - 1) mod $n
};

(:
 : twixt()
 : Return true of the value if between the minimum and maximum.
 :
 : $value: Value to test
 : $min: Minimum (inclusive)
 : $max: Maximum (inclusive)
 :)
declare function this:twixt($value as xs:double, $min as xs:double?, $max as xs:double?)
{
  (if (exists($min)) then ($value ge $min) else true()) and
  (if (exists($max)) then ($value le $max) else true())
};

(:
 : clamp()
 : Force the value into the range by mapping high values to the maximum
 : and low values to the minimum.
 :
 : $value: Value to clamp
 : $min: Minimum (inclusive)
 : $max: Maximum (inclusive)
 :)
declare function this:clamp($value as xs:double, $min as xs:double?, $max as xs:double?)
{
  if (exists($min) and $value lt $min) then $min
  else if (exists($max) and $value gt $max) then $max
  else $value
};

(:
 : sign()
 : Sign of value: -1 if < 0, 0 if 0, +1 if > 0
 :)
declare function this:sign($v as xs:double) as xs:double
{
  if ($v = 0) then 0
  else if ($v < 0) then -1
  else 1
};

(:
 : smoothstep()
 : Hermite interpolation with edge clamping
 : 0 if less than low bound, 1 if more than high bound, 3x² - 2x³ between
 :)
declare function this:smoothstep(
  $low as xs:double,
  $high as xs:double,
  $v as xs:double
) as xs:double
{
  if ($low > $high) then errors:error("ML-BADARGS", ("low", $low)) else (),
  let $x := this:clamp(($v - $low) div ($high - $low), 0.0, 1.0)
  return $x * $x * (3 - 2 * $x)
};

(:
 : mix()
 : Linear combination of two numbers
 :)
declare function this:mix(
  $a as xs:double,
  $b as xs:double,
  $f as xs:double
) as xs:double
{
  $a + ($b - $a) * $f  
};

declare variable $this:RADIANS_PER_DEGREE as xs:double := math:pi() div 180;
declare variable $this:DEGREES_PER_RADIAN as xs:double := 180 div math:pi();

(:
 : radians()
 : Degrees to radians conversion
 :)
declare function this:radians($degrees as xs:double) as xs:double
{
  $this:RADIANS_PER_DEGREE * $degrees
};

(:
 : degrees()
 : Radians to degrees conversion
 :)
declare function this:degrees($radians as xs:double) as xs:double
{
  $this:DEGREES_PER_RADIAN * $radians
};

(:======================================================================
 : Sequence operations
 :======================================================================:)

(:
 : linspace()
 : Return a sequence of evenly (linearly) spaced values between two values.
 : The lower bound is included, the upper bound is unless the exlusive flag
 : is set.
 : 
 : e.g. 
 : linspace(5, 2, 3, false()) => (2, 2.25, 2.5, 2.75, 3)
 : linspace(5, 2, 3, true())  => (2, 2.2, 2.4, 2.6, 2.8) 
 :
 : $n: number of values to return
 : $from: starting value
 : $to: ending value
 : $exclusive: whether to include the upper bound or not
 :)
declare function this:linspace(
  $n as xs:integer,
  $from as xs:double,
  $to as xs:double,
  $exclusive as xs:boolean
) as xs:double*
{
  if ($n = 1) then (
    $from,
    if ($exclusive) then () else $to
  ) else (
    let $space := 
      if ($exclusive)
      then ($to - $from) div $n
      else ($to - $from) div ($n - 1)
    for $i in 1 to $n
    return $from + ($i - 1)*$space
  )
};


(:
 : linspace()
 : Return a sequence of evenly (linearly) spaced values between two values.
 : The bounds are included.
 :
 : $n: number of values to return
 : $from: starting value
 : $to: ending value
 :)
declare function this:linspace(
  $n as xs:integer,
  $from as xs:double,
  $to as xs:double
) as xs:double*
{
  this:linspace($n, $from, $to, false())
};

(:
 : powspace()
 : Return numbers between 0 and 1 spread out according to math:pow($fade, $i)
 : 
 : $n: number of interpolations
 : $fade: the power fade
 :   >1 escalating gaps towards 1
 :   <1 shrinking gaps towards 1
 : $exclusive: true iff we omit 1
 :)
declare function this:powspace(
  $n as xs:integer,
  $fade as xs:double,
  $exclusive as xs:boolean
) as xs:double*
{
  if ($fade = 1) then (
    this:linspace($n, 0, 1, $exclusive)
  ) else (
    for $t in this:linspace($n, 0, 1, $exclusive)
    return math:pow($t, $fade)
  )
};

(:
 : powspace()
 : Return numbers between 0 and 1 (inclusive) spread out according to
 : math:pow($fade, $i)
 : 
 : $n: number of interpolations
 : $fade: the power fade
 :   >1 escalating gaps towards 1
 :   <1 shrinking gaps towards 1
 :)
declare function this:powspace(
  $n as xs:integer,
  $fade as xs:double
) as xs:double*
{
  this:assert($fade > 0, "Fade must be positive"),
  this:powspace($n, $fade, false())
};

(:
 : arange()
 : Return a sequence of numbers starting at the lower bound and ending
 : before the upper bound, separated by a given step.
 :
 : arange(3, 7, 2) => 3, 5
 : 
 : $from: lower bound
 : $to: upper bound
 : $step: difference between adjacent numbers
 :)
declare function this:arange(
  $from as xs:double,
  $to as xs:double,
  $step as xs:double
) as xs:double*
{
  let $n := ceiling(($to - $from) div $step) cast as xs:integer
  for $i in 1 to $n
  return $from + ($i - 1)*$step
};

(: 
 : zip()
 : Map a function over two sequences in parallel.
 : Sequences should be the same size, but if they aren't we'll get a
 : result sequence the same size as the shorter sequence.
 :)
declare function this:zip(
  $f as function(item(), item()) as item(),
  $seq1 as item()*,
  $seq2 as item()*
) as item()*
{
  (: if |seq2| > |seq1| we just skip the tail of seq2 :)
  (: if |seq1| < |seq2| avoid calling $f on the excess values in seq1 :)
  for $v1 at $i in $seq1 return (
    if (empty($seq2[$i])) then () else $f($v1, $seq2[$i])
  )
};

(:
 : min-key()
 : Return the value in the collection where $f($value) is minimum.
 : Differs from map-min-key(map-entries()) if entries are sequences.
 :
 : $f: function over values to minimize
 : $collection: values
 :)
declare function this:min-key(
  $f as function(item()*) as xs:double,
  $collection as item()*
) as item()?
{
  (for $item in $collection
   order by $f($item) ascending
   return $item)[1]
};

declare function this:min-key(
  $collection as item()*
) as item()?
{
  (for $item in $collection
   order by $item ascending
   return $item)[1]
};

(:
 : max-key()
 : Return the value in the collection where $f($value) is maximum.
 : Differs from map-max-key(map-entries()) if entries are sequences.
 :
 : $f: function over values to maximize
 : $collection: values
 :)
declare function this:max-key(
  $f as function(item()*) as xs:double,
  $collection as item()*
) as item()?
{
  (for $item in $collection
   order by $f($item) descending
   return $item)[1]
};

declare function this:max-key(
  $collection as item()*
) as item()?
{
  (for $item in $collection
   order by $item descending
   return $item)[1]
};

(:====================================================================== 
 : "Loops"
 :======================================================================:)

(: 
 : while()
 : Execute body repeatedly while predicate is true
 :
 : $predicate: boolean test over $data
 : $body: operation to execute over $data
 : $data: value
 :)
declare function this:while(
  $predicate as function(item()*) as xs:boolean,
  $body as function(item()*) as item()*,
  $data as item()*)
{
  if (not($predicate($data))) then $data else (
    this:while($predicate, $body, $body($data))
  )
};

(:
 : while()
 : Execute the body repeatedly while the predicate is true, with
 : automatic marshalling and unmarshalling of a data sequence of 2 values
 : Caller will have to unmarshal the values from the result.
 : Note: requires first value to be singleton
 :)
declare function this:while(
  $predicate as function(item(),item()*) as xs:boolean,
  $body as function(item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item()*
)
{
  if (not($predicate($v1,$v2))) then ($v1,$v2) else (
    let $data := $body($v1,$v2)
    return this:while($predicate, $body, head($data), tail($data))
  )
};

(:
 : while()
 : Execute the body repeatedly while the predicate is true, with
 : automatic marshalling and unmarshalling of a data sequence of 3 values
 : Caller will have to unmarshal the values from the result.
 : Note: requires first 2 values to be singletons
 :)
declare function this:while(
  $predicate as function(item(),item(),item()*) as xs:boolean,
  $body as function(item(),item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item(),
  $v3 as item()*
)
{
  if (not($predicate($v1,$v2,$v3))) then ($v1,$v2,$v3) else (
    let $data := $body($v1,$v2,$v3)
    return this:while($predicate, $body, head($data), head(tail($data)), tail(tail($data)))
  )
};

(:
 : while()
 : Execute the body repeatedly while the predicate is true, with
 : automatic marshalling and unmarshalling of a data sequence of 4 values
 : Caller will have to unmarshal the values from the result.
 : Note: requires first 3 values to be singletons
 :)
declare function this:while(
  $predicate as function(item(),item(),item(),item()*) as xs:boolean,
  $body as function(item(),item(),item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item(),
  $v3 as item(),
  $v4 as item()*
)
{
  if (not($predicate($v1,$v2,$v3,$v4))) then ($v1,$v2,$v3,$v4) else (
    let $data := $body($v1,$v2,$v3,$v4)
    return
      this:while($predicate, $body,
        head($data),
        head(tail($data)),
        head(tail(tail($data))),
        tail(tail(tail($data)))
      )
  )
};

(:
 : while()
 : Execute the body repeatedly while the predicate is true, with
 : automatic marshalling and unmarshalling of a data sequence of 5 values
 : Caller will have to unmarshal the values from the result.
 : Note: requires first 4 values to be singletons
 :)
declare function this:while(
  $predicate as function(item(),item(),item(),item(),item()*) as xs:boolean,
  $body as function(item(),item(),item(),item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item(),
  $v3 as item(),
  $v4 as item(),
  $v5 as item()*
)
{
  if (not($predicate($v1,$v2,$v3,$v4,$v5))) then ($v1,$v2,$v3,$v4,$v5) else (
    let $data := $body($v1,$v2,$v3,$v4,$v5)
    return
      this:while($predicate, $body,
        head($data),
        head(tail($data)),
        head(tail(tail($data))),
        head(tail(tail(tail($data)))),
        tail(tail(tail(tail($data))))
      )
  )
};

(:
 : while()
 : Execute the body repeatedly while the predicate is true, with
 : automatic marshalling and unmarshalling of a data sequence of 6 values
 : Caller will have to unmarshal the values from the result.
 : Note: requires first 5 values to be singletons
 :)
declare function this:while(
  $predicate as function(item(),item(),item(),item(),item(),item()*) as xs:boolean,
  $body as function(item(),item(),item(),item(),item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item(),
  $v3 as item(),
  $v4 as item(),
  $v5 as item(),
  $v6 as item()*
)
{
  if (not($predicate($v1,$v2,$v3,$v4,$v5,$v6))) then ($v1,$v2,$v3,$v4,$v5,$v6) else (
    let $data := $body($v1,$v2,$v3,$v4,$v5,$v6)
    return
      this:while($predicate, $body,
        head($data),
        head(tail($data)),
        head(tail(tail($data))),
        head(tail(tail(tail($data)))),
        head(tail(tail(tail(tail($data))))),
        tail(tail(tail(tail(tail($data)))))
      )
  )
};

(: 
 : until()
 : Execute body repeatedly until predicate is true
 : Warning: could run forever if predicate never becomes true
 :
 : $predicate: boolean test over $data
 : $body: operation to execute over $data
 : $data: value
 :)
declare function this:until(
  $predicate as function(item()*) as xs:boolean,
  $body as function(item()*) as item()*,
  $data as item()*)
{
  let $new-data := $body($data)
  return
    if ($predicate($new-data)) then $new-data
    else this:until($predicate, $body, $new-data)
};

(: 
 : until()
 : Execute body repeatedly until predicate is true, with automatic
 : marshalling and unmarshalling of a data sequence of 2 values.
 : Caller will have to unmarshal the values from the result.
 : Note: requires first value to be singleton
 : Warning: could run forever if predicate never becomes true
 :)
declare function this:until(
  $predicate as function(item(),item()*) as xs:boolean,
  $body as function(item(),item()*) as item()*,
  $v1 as item(),
  $v2 as item()*
)
{
  let $new-data := $body($v1,$v2)
  return
    if ($predicate(head($new-data), tail($new-data))) then $new-data
    else this:until($predicate, $body, head($new-data), tail($new-data))
};

(:====================================================================== 
 : Tree walks
 :======================================================================:)

(:
 : depth-first-preorder()
 : Depth first traversal of tree structure, where we take action on the
 : node before walking the children.
 : 
 : $node: something representing the current node
 : $action: function to act on a node as it is traversed
 : $children: function that returns a sequence of children of a node
 : $level: current level in the walk (start at 0)
 :)
declare function this:depth-first-preorder(
  $node as item(),
  $action as function(item(), xs:integer) as item()*,
  $children as function(item()) as item()*,
  $level as xs:integer
) as item()*
{
  $action($node, $level),
  for $child in $children($node)
  return this:depth-first-preorder($child, $action, $children, $level + 1)
};

(:
 : depth-first-postorder()
 : Depth first traversal of tree structure, where we take action on the
 : node after walking the children.
 : 
 : $node: something representing the current node
 : $action: function to act on a node as it is traversed
 : $children: function that returns a sequence of children of a node
 : $level: current level in the walk (start at 0)
 :)
declare function this:depth-first-postorder(
  $node as item(),
  $action as function(item(), xs:integer) as item()*,
  $children as function(item()) as item()*,
  $level as xs:integer
) as item()*
{
  for $child in $children($node)
  return this:depth-first-postorder($child, $action, $children, $level + 1),
  $action($node, $level)
};

(:
 : breadth-first-preorder()
 : Breadth first traversal of tree structure, where we take action on each
 : node and its siblings before walking them.
 : 
 : $node: something representing the current node
 : $action: function to act on a node as it is traversed
 : $children: function that returns a sequence of children of a node
 : $level: current level in the walk (start at 0)
 :)
declare function this:breadth-first-preorder(
  $node as item(),
  $action as function(item(), xs:integer) as item()*,
  $children as function(item()) as item()*,
  $level as xs:integer
) as item()*
{
  if ($level=0) then $action($node, $level) else (),
  for $child in $children($node) return $action($child,$level+1),
  for $child in $children($node)
  return this:breadth-first-preorder($child,$action,$children,$level+1)
};

(:
 : breadth-first-postorder()
 : Breadth first traversal of tree structure, where we take action on each
 : node and its siblings after walking them.
 : 
 : $node: something representing the current node
 : $action: function to act on a node as it is traversed
 : $children: function that returns a sequence of children of a node
 : $level: current level in the walk (start at 0)
 :)
declare function this:breadth-first-postorder(
  $node as item(),
  $action as function(item(), xs:integer) as item()*,
  $children as function(item()) as item()*,
  $level as xs:integer
) as item()*
{
  for $child in $children($node)
  return this:breadth-first-postorder($child,$action,$children,$level+1),
  for $child in $children($node) return $action($child,$level+1),
  if ($level=0) then $action($node, $level) else ()
};

(:====================================================================== 
 : Map operations
 :======================================================================:)

(:
 : merge-into()
 : Create a map formed by replacing entries with keys matching new keys
 : with the new values, while preserving unmatched entries from old map
 :
 : e.g. merge-into({a:1, b:1}, {a:2, c:2}) = {a:2, b:1, c:2})
 : e.g. merge-into({a:2, c:2}, {a:1, b:1}) = {a:1, b:1, c:2})
 : 
 : $old: old map
 : $new: new map
 :)
declare function this:merge-into($old as map(*), $new as map(*)) as map(*)
{
  if (empty($new)) then $old
  else map:merge(($old,$new), map { "duplicates" : "use-last" })
};

(:
 : merge-into()
 : Merge into each map in order to create a map formed by replacing entries
 : with keys matching new keys with the new values, while preserving unmatched 
 : entries from earlier maps in the sequence
 :
 : Like XQuery 3.1 map:merge($maps, map {"duplicates" : "use-last" })
 :
 : $maps: sequence of maps
 :)
declare function this:merge-into($maps as map(*)*) as map(*)
{
  map:merge($maps, map { "duplicates" : "use-last" })
};

(:
 : exclude()
 : Create a new map formed by removing entries with keys matching new keys
 : while preserving unmatched entries from old map
 :
 : e.g. exclude({a:1, b:1}, ("a", "c") = {b:1})
 :
 : $old: old map
 : $excludes: keys to delete
 :)
declare function this:exclude($old as map(*), $excludes as xs:string*) as map(*)
{
  fold-left(
    $excludes,
    $old,
    function($map as map(*), $exclude) as map(*) {
      $map=>map:remove($exclude)
    }
  )
};

(:
 : map-append()
 : Append a value to the existing value for the given key; return new map
 : e.g. {"a": (1,2)}=>util:map-append("a", 3) => {"a": (1,2,3)}
 : 
 : $map: the map
 : $key: the key
 : $value: the value to append
 :)
declare function this:map-append(
  $map as map(*),
  $key as xs:anyAtomicType,
  $value as item()
) as map(*)
{
  $map=>map:put($key, ($map=>map:get($key), $value))
};

(:
 : map-increment()
 : Increment the existing value for the given key; return new map
 : e.g. {"a": 5}=>util:map-increment("a") => {"a": 6}
 : 
 : $map: the map
 : $key: the key
 : $value: the value to append
 :)
declare function this:map-increment(
  $map as map(*),
  $key as xs:anyAtomicType
) as map(*)
{
  $map=>map:put($key, (($map=>map:get($key),0)[1] + 1))
};

(:
 : map-entries()
 : Return all the values in the map
 : 
 : $map: the map
 :)
declare function this:map-entries($map as map(*)) as item()*
{
  for $key in $map=>map:keys() return $map=>map:get($key)
};

(:
 : map-deconstruct()
 : Convert each entry in map into its own mini-map.
 : 
 : Example:
 : map-deconstruct(map {"a":1, "b":2}) => {"a":1} {"b":2}
 :)
declare function this:map-deconstruct($map as map(*)) as map(*)*
{
  for $key in $map=>map:keys()
  return map { $key : $map=>map:get($key) }
};

(:
 : map-construct()
 : Convert separate map enties into a single map (inverse of map-deconstruct())
 : 
 : Example:
 : map-construct((map {"a":1}, map {"b":2}) => {"a":1, "b":2}
 : Multiple keys get merged:
 : map-construct((map {"a", 1}, map {"a": 2})) => {"a": (1,2)}
 :)
declare function this:map-construct($maps as map(*)*) as map(*)
{
  map:merge($maps, map { "duplicates" : "combine" })
};

(:
 : map-keys-of-min()
 : Return the keys of the entries with the minimum value
 :
 : $map: the map
 :)
declare function this:map-keys-of-min($map as map(*)) as xs:anyAtomicType*
{
  let $min := min(this:map-entries($map))
  return ($map=>map:keys())[$map=>map:get(.)=$min]
};

(:
 : map-keys-of-max()
 : Return the keys of the entries with the maximum value
 :
 : $map: the map
 :)
declare function this:map-keys-of-max($map as map(*)) as xs:anyAtomicType*
{
  let $max := max(this:map-entries($map))
  return ($map=>map:keys())[$map=>map:get(.)=$max]
};

(:
 : map-min-key()
 : Return the entry in the map where $f($entry) is minimum.
 :
 : $f: function over values to minimize
 : $map: the map
 :)
declare function this:map-min-key(
  $f as function(item()*) as xs:double,
  $map as map(*)
) as item()*
{
  $map=>map:get(
    (for $key in $map=>map:keys()
     order by $f($map=>map:get($key)) ascending
     return $key)[1]
  )
};

(:
 : map-max-key()
 : Return the entry in the map where $f($entry) is maximum.
 :
 : $f: function over values to maximize
 : $map: the map
 :)
declare function this:map-max-key(
  $f as function(item()*) as xs:double,
  $map as map(*)
) as item()*
{
  $map=>map:get(
    (for $key in $map=>map:keys()
     order by $f($map=>map:get($key)) descending
     return $key)[1]
  )
};

(:====================================================================== 
 : Misc
 :======================================================================:)

declare function this:hex-to-integer($hex as xs:string) as xs:integer
{
  if (string-length($hex) gt 16) then errors:error("UTIL-BADHEX", $hex)
  else (
    sum(
      for $cp at $i in reverse(string-to-codepoints($hex)) return (
        if ($cp=(48,49,50,51,52,53,54,55,56,57)) (: 0 to 9 :)
        then ($cp - 48)*math:pow(16, $i - 1)
        else if ($cp=(97,98,99,100,101,102)) (: a to f :)
        then ($cp - 97 + 10)*math:pow(16, $i - 1)
        else if ($cp=(65,66,67,68,69,70)) (: A to F :)
        then ($cp - 65 + 10)*math:pow(16, $i - 1)
        else errors:error("UTIL-BADHEX", $hex)
      )
    ) cast as xs:integer
  )
};

declare function this:repeat($n as xs:integer, $val as item()) as item()*
{
  (1 to $n)!$val
};

(:
 : describe-qname()
 : Clark notation version of a QName, suitable for dumping out function
 : names in a random distribution, for example.
 :
 : $qname: the QName
 :)
declare function this:describe-qname($qname as xs:QName) as xs:string
{
  "{"||namespace-uri-from-QName($qname)||"}"||local-name-from-QName($qname)
};

(:
 : assert()
 : Raise an error if the fact is not true.
 :)
declare function this:assert($that as xs:boolean, $complaint as xs:string) as empty-sequence()
{
  if ($that) then ()
  else error(QName("http://mathling.com/errors", "ml:ASSERTFAIL"), $complaint)
};

(: utilities.xqy :)
