xquery version "3.1";
(:
 : Edges: including arcs, quads, and cubics
 : Edges are decorated with properties which art infrastructure may care
 : about for rendering purposes.
 : 
 : Properties of note:
 :   r: radius of arc for arc edges; default is none
 :   width: width for stroke; default=d
 :   label: label for this edge
 :   class, colour, opacity: direct styling overrides
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/geometric/edge"; 

import module namespace errors="http://mathling.com/core/errors"
       at "../core/errors.xqy";
import module namespace util="http://mathling.com/core/utilities"
       at "../core/utilities.xqy";
import module namespace point="http://mathling.com/geometric/point"
       at "point.xqy";
import module namespace ellipse="http://mathling.com/geometric/ellipse"
       at "ellipse.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

declare variable $this:EDGE-RESERVED as xs:string* := ("kind", "u", "v", "d");

(:
 : edge()
 : Make a graph edge structure
 :
 : $from: starting point
 : $to: ending point
 : $d: weight of the edge (used by spanning graph as distance)
 : $properties: additional graph properties
 :)
declare function this:edge(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $d as xs:double,
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  util:merge-into($properties,
    map {
      "kind": "edge",
      "u": $from,
      "v": $to,
      "d": $d
    }
  )
};

(:
 : edge()
 : Make a graph edge structure
 : 
 : $from: starting point
 : $to: ending point
 : $d: the weight of the edge
 :)
declare function this:edge(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $d as xs:double
) as map(xs:string,item()*)
{
  map {
    "kind": "edge",
    "u": $from,
    "v": $to,
    "d": $d
  }
};

(:
 : edge()
 : Make a graph edge structure
 : 
 : $from: starting point
 : $to: ending point
 :)
declare function this:edge(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:edge($from, $to, 0)
};

declare function this:kind($edge as map(xs:string,item()*)) as xs:string
{
  $edge("kind")
};

(:
 : start()
 : Accessor for starting point of edge
 : Handles graph and polygon edges
 :   $edge: The edge
 :)
declare function this:start($edge as map(xs:string,item()*)) as map(xs:string,item()*)
{
  switch (this:kind($edge))
  case "edge" return $edge("u")
  case "quad" return $edge("u")
  case "cubic" return $edge("u")
  case "arc" return (
    let $start := $edge("start")
    return
      if (exists($start)) then $start
      else (
        let $circle := $edge("circle")
        let $center := ellipse:center($circle)
        let $r := ellipse:radius($circle)
        let $from := util:radians($edge("from-angle"))
        return (
          point:point(
            point:px($center)+math:cos($from)*$r,
            point:py($center)+math:sin($from)*$r
          )
        )
      )
    )
  default return errors:error("GEOM-BADREGION", $edge)
};

(:
 : end()
 : Accessor for ending point of edge
 : Handles graph and polygon edges
 :   $edge: The edge
 :)
declare function this:end($edge as map(xs:string,item()*)) as map(xs:string,item()*)
{
  switch (this:kind($edge))
  case "edge" return $edge("v")
  case "quad" return $edge("v")
  case "cubic" return $edge("v")
  case "arc" return (
    let $end := $edge("end")
    return
      if (exists($end)) then $end
      else (
        let $circle := $edge("circle")
        let $center := ellipse:center($circle)
        let $r := ellipse:radius($circle)
        let $to := util:radians($edge("to-angle"))
        return 
          point:point(
            point:px($center)+math:cos($to)*$r,
            point:py($center)+math:sin($to)*$r
          )
      )
  )
  default return errors:error("GEOM-BADREGION", $edge)
};

(:
 : weight()
 : Accessor for the weight of a graph edge
 : Handles graph and polygon edges (polygon edges are always weight 0)
 :   $edge: The edge
 :)
declare function this:weight($edge as map(xs:string,item()*)) as xs:double
{
  if (this:kind($edge) = ("edge","cubic","quad"))
  then ($edge("d"),0)[1]
  else 0
};

declare variable $this:ARC-RESERVED as xs:string* :=
  ("kind", "circle", "start", "end", "from-angle", "to-angle", "flipped", "large")
;

(:
 : arc()
 : Construct an arc (e.g. hyperbolic edge)
 : Note: calculations on $start and $end need to be correct: 
 : use hyperbolic:arc-through() to get them right; this is really an
 : internal contructor function. Ad hoc construction of arc should use 
 : arc-by-angle()
 :   $center: center of curvature
 :   $radius: radius of circle
 :   $start: starting point
 :   $end: ending point
 :   $flipped: indicates sweep
 :   $large: indicated large arc
 :)
declare function this:arc(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $start as map(xs:string,item()*),
  $end as map(xs:string,item()*),
  $flipped as xs:boolean,
  $large as xs:boolean
) as map(xs:string,item()*)
{
  map {
    "kind": "arc",
    "circle": ellipse:circle($center, $radius),
    "start": $start,
    "end": $end,
    "flipped": $flipped,
    "large": $large
  }
};

declare function this:arc(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $start as map(xs:string,item()*),
  $end as map(xs:string,item()*),
  $flipped as xs:boolean
) as map(xs:string,item())*
{
  this:arc($center, $radius, $start, $end, $flipped, false())
};

declare function this:arc(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $start as map(xs:string,item()*),
  $end as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:arc($center, $radius, $start, $end, false(), false())
};

(:
 : arc-by-angle()
 : Construct an arc
 :   $center: center of curvature
 :   $radius: radius of circle
 :   $θ: starting angle (degrees)
 :   $θ2: ending angle (degrees)
 :   $flipped: indicates sweep
 :   $large: indicates large arc
 :)
declare function this:arc-by-angle(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $θ as xs:double,
  $θ2 as xs:double,
  $flipped as xs:boolean,
  $large as xs:boolean
) as map(xs:string,item()*)
{
  map {
    "kind": "arc",
    "circle": ellipse:circle($center, $radius),
    "from-angle": $θ,
    "to-angle": $θ2,
    "flipped": $flipped,
    "large": $large
  }
};

declare function this:arc-by-angle(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $θ as xs:double,
  $θ2 as xs:double,
  $flipped as xs:boolean
) as map(xs:string,item()*)
{
  this:arc-by-angle($center, $radius, $θ, $θ2, $flipped, false())
};

declare function this:arc-by-angle(
  $center as map(xs:string,item()*),
  $radius as xs:double,
  $θ as xs:double,
  $θ2 as xs:double
) as map(xs:string,item()*)
{
  this:arc-by-angle($center, $radius, $θ, $θ2, false(), false())
};

(: 
 : arc-circle()
 : Center and radius of curvature for this arc
 :   $arc: The arc
 :)
declare function this:arc-circle($arc as map(xs:string,item()*)) as map(xs:string,item()*)
{
  $arc("circle")
};

(: 
 : arc-center()
 : Center of this arc
 :   $arc: The arc
 :)
declare function this:arc-center($arc as map(xs:string,item()*)) as map(xs:string,item()*)
{
  $arc("circle")=>ellipse:center()
};

(: 
 : arc-radius()
 : Radius of curvature of this arc
 :   $arc: The arc
 :)
declare function this:arc-radius($arc as map(xs:string,item()*)) as xs:double
{
  $arc("circle")=>ellipse:radius()
};

(: 
 : arc-angles()
 : Starting and ending angles for the arc (degrees), for arc-by-angle arcs
 :   $arc: The arc
 :)
declare function this:arc-angles($arc as map(xs:string,item()*)) as xs:double*
{
  ($arc("from-angle"),$arc("to-angle"))
};

(: 
 : arc-ends()
 : Starting and ending points for the arc, for arc-by-point arcs
 :   $arc: The arc
 :)
declare function this:arc-ends($arc as map(xs:string,item()*)) as map(xs:string,item()*)*
{
  ($arc("start"),$arc("end"))
};

(: 
 : arc-flipped()
 : Sweep for the arc
 :   $arc: The arc
 :)
declare function this:arc-flipped($arc as map(xs:string,item()*)) as xs:boolean
{
  (: Default is false :)
  if ($arc("flipped")) then true() else false()
};

declare function this:arc-large($arc as map(xs:string,item()*)) as xs:boolean
{
  (: Default is false :)
  if ($arc("large")) then true() else false()
};

declare variable $this:QUAD-RESERVED as xs:string* :=
  ($this:EDGE-RESERVED, "c1");

(:
 : quad()
 : Construct a quadratic edge.
 : 
 : $from: starting point
 : $to: ending point
 : $control: control point
 : $d: edge weight
 : $properties: additional properties
 :)
declare function this:quad(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control as map(xs:string,item()*),
  $d as xs:double,
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  util:merge-into($properties,
    map {
      "kind": "quad",
      "u": $from,
      "v": $to,
      "c1": $control,
      "d": $d
    }
  )
};

declare function this:quad(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control as map(xs:string,item()*),
  $d as xs:double
) as map(xs:string,item()*)
{
  map {
    "kind": "quad",
    "u": $from,
    "v": $to,
    "c1": $control,
    "d": $d
  }
};


declare function this:quad(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:quad($from, $to, $control, 0)
};

(:
 : controls()
 : Get control points from a quadratic or cubic edge.
 :)
declare function this:controls(
  $edge as map(xs:string,item()*)
) as map(xs:string,item()*)*
{
  switch(this:kind($edge))
  case "quad" return $edge("c1")
  case "cubic" return ($edge("c1"),$edge("c2"))
  default return errors:error("GEOM-BADREGION", $edge)
};

declare variable $this:CUBIC-RESERVED as xs:string* :=
  ($this:EDGE-RESERVED, "c1", "c2");

(:
 : cubic()
 : Construct a cubic edge.
 : 
 : $from: starting point
 : $to: ending point
 : $control1: first control point
 : $control2: second control point
 : $d: edge weight
 : $properties: additional properties
 :)
declare function this:cubic(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control1 as map(xs:string,item()*),
  $control2 as map(xs:string,item()*),
  $d as xs:double,
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  util:merge-into($properties,
    map {
      "kind": "cubic",
      "u": $from,
      "v": $to,
      "c1": $control1,
      "c2": $control2,
      "d": $d
    }
  )
};

declare function this:cubic(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control1 as map(xs:string,item()*),
  $control2 as map(xs:string,item()*),
  $d as xs:double
) as map(xs:string,item()*)
{
  map {
    "kind": "cubic",
    "u": $from,
    "v": $to,
    "c1": $control1,
    "c2": $control2,
    "d": $d
  }
};

declare function this:cubic(
  $from as map(xs:string,item()*),
  $to as map(xs:string,item()*),
  $control1 as map(xs:string,item()*),
  $control2 as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:cubic($from, $to, $control1, $control2, 0)
};

(:======================================================================
 : Property management
 :======================================================================:)
 
(:
 : property-map()
 : Return the annotation properties of the edge as a map. Check whether this
 : is actually an edge.
 : 
 : $region: the region
 :)
declare function this:property-map(
  $region as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  switch(this:kind($region))
  case "edge" return util:exclude($region,$this:EDGE-RESERVED)
  case "arc" return util:exclude($region,$this:ARC-RESERVED)
  case "quad" return util:exclude($region, $this:QUAD-RESERVED)
  case "cubic" return util:exclude($region, $this:CUBIC-RESERVED)
  default return map {}
};

(:
 : properties()
 : Return the names of the annotation properties of the edge.
 : Check whether this is actually an edge
 : 
 : $region: the region
 :)
declare function this:properties(
  $region as map(xs:string,item()*)
) as xs:string*
{
  switch(this:kind($region))
  case "edge" return ($region=>map:keys())[not(. = $this:EDGE-RESERVED)]
  case "arc" return ($region=>map:keys())[not(. = $this:ARC-RESERVED)]
  case "quad" return ($region=>map:keys())[not(. = $this:QUAD-RESERVED)]
  case "cubic" return ($region=>map:keys())[not(. = $this:CUBIC-RESERVED)]
  default return ()
};

(:
 : with-properties()
 : Annotate the edge with some new properties and return the new edge.
 : Will not touch any of the core properties. Will override existing properties
 : with the same keys but leave properties with different keys in place.
 : Raises an error if this is not actually an edge.
 : 
 : $region: the region
 :)
declare function this:with-properties(
  $region as map(xs:string,item()*),
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  switch(this:kind($region))
  case "edge" return util:merge-into($region, util:exclude($properties,$this:EDGE-RESERVED))
  case "arc" return util:merge-into($region, util:exclude($properties,$this:ARC-RESERVED))
  case "quad" return util:merge-into($region, util:exclude($properties, $this:QUAD-RESERVED))
  case "cubic" return util:merge-into($region, util:exclude($properties, $this:CUBIC-RESERVED))
  default return errors:error("GEOM-BADREGION", $region)
};

(:======================================================================
 : Operations
 :======================================================================:)
 
(:
 : snap()
 : Snap the coordinates of the points, returning the edges with snapped
 : (i.e. integer) points coordinates.
 :
 : $edges the edges
 :)
declare function this:snap(
  $edges as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  for $edge in $edges return switch(this:kind($edge))
  case "edge" return
    this:edge(
      point:snap(this:start($edge)),
      point:snap(this:end($edge)),
      this:weight($edge),
      $edge
    )
  case "quad" return
    this:quad(
      point:snap(this:start($edge)),
      point:snap(this:end($edge)),
      point:snap(this:controls($edge)[1]),
      this:weight($edge),
      $edge
    )
  case "cubic" return
    let $controls := this:controls($edge)
    return
      this:cubic(
        point:snap(this:start($edge)),
        point:snap(this:end($edge)),
        point:snap($controls[1]),
        point:snap($controls[2]),
        this:weight($edge),
        $edge
      )
  case "arc" return
    let $circle := $edge=>this:arc-circle()
    let $points := $edge=>this:arc-ends()
    let $angles := $edge=>this:arc-angles()
    return
      if (empty($points)) then (
        util:assert(exists($angles),"Arc must have points or angles"),
        $edge=>
          map:put("circle", ellipse:snap($circle))
      ) else (
        $edge=>
           map:put("circle", ellipse:snap($circle))=>
           map:put("start", point:snap($points[1]))=>
           map:put("end", point:snap($points[2]))
      )
  default return $edge
};

(:
 : decimal()
 : Perform decimal rounding on all the point coordinates (see util:decimal).
 :
 : $edges: the edges to round
 : $digits: how many digits after the decimal point to keep
 :)
declare function this:decimal(
  $edges as map(xs:string,item()*)*,
  $digits as xs:integer
) as map(xs:string,item()*)*
{
  for $edge in $edges return switch(this:kind($edge))
  case "edge" return
    this:edge(
      point:decimal(this:start($edge), $digits),
      point:decimal(this:end($edge), $digits),
      this:weight($edge),
      $edge
    )
  case "quad" return
    this:quad(
      point:decimal(this:start($edge), $digits),
      point:decimal(this:end($edge), $digits),
      point:decimal(this:controls($edge)[1], $digits),
      this:weight($edge),
      $edge
    )
  case "cubic" return
    let $controls := this:controls($edge)
    return
      this:cubic(
        point:decimal(this:start($edge), $digits),
        point:decimal(this:end($edge), $digits),
        point:decimal($controls[1], $digits),
        point:decimal($controls[2], $digits),
        this:weight($edge),
        $edge
      )
  case "arc" return
    let $circle := $edge=>this:arc-circle()
    let $points := $edge=>this:arc-ends()
    let $angles := $edge=>this:arc-angles()
    return
      if (empty($points)) then (
        util:assert(exists($angles),"Arc must have points or angles"),
        $edge=>
          map:put("circle", ellipse:decimal($circle, $digits))
      ) else (
        $edge=>
           map:put("circle", ellipse:decimal($circle, $digits))=>
           map:put("start", point:decimal($points[1], $digits))=>
           map:put("end", point:decimal($points[2], $digits))
      )
  default return $edge
};

(:
 : quote()
 : Return a string value for the edges, suitable for debugging.
 : 
 : $edges: the edge sequence to quote
 :)
declare function this:quote(
  $edges as map(xs:string,item()*)*
) as xs:string
{
  string-join(
    for $edge in $edges return switch(this:kind($edge))
    case "edge" return
      point:quote($edge("u"))||":"||point:quote($edge("v"))
    case "quad" return
      point:quote($edge("u"))||":"||point:quote($edge("v"))||"<"||point:quote($edge("c1"))||">"
    case "cubic" return
      point:quote($edge("u"))||":"||point:quote($edge("v"))||"<"||point:quote($edge("c1"))||";"||point:quote($edge("c2"))||">"
    case "arc" return
      "("||
        ellipse:quote($edge("circle"))||
        "<"||
          point:quote($edge("start"))||
          $edge("from-angle")||":"||
          point:quote($edge("end"))||
          $edge("to-angle")||
        ">"||
        (if ($edge("large")) then "1" else "0")||
        (if ($edge("flipped")) then "1" else "0")||
      ")"
    default return errors:quote($edge)
    ,
    " "
  )
};

(:
 : same()
 : Equality comparison for edges, ignoring annotation properties.
 : Return true() if they have equal coordinates.
 : 
 : $this: one edge
 : $other: the edge to compare it to
 :)
declare function this:same(
  $this as map(xs:string,item()*),
  $other as map(xs:string,item()*)
) as xs:boolean
{
  let $this-kind := this:kind($this)
  let $other-kind := this:kind($other)
  return
  (
    ($this-kind=$other-kind) and (
    switch($this-kind)
    case "arc" return
      ellipse:same($this("circle"),$other("circle")) and
      (
        if (empty($this("start"))) then empty($other("start"))
        else if (empty($other("start"))) then false()
        else point:same($this("start"), $other("start"))
      ) and
      (
        if (empty($this("end"))) then empty($other("end"))
        else if (empty($other("end"))) then false()
        else point:same($this("end"), $other("end"))
      ) and (
        if (empty($this("from-angle"))) then empty($other("from-angle"))
        else if (empty($other("from-angle"))) then false()
        else $this("from-angle")=$other("from-angle")
      ) and (
        if (empty($this("to-angle"))) then empty($other("to-angle"))
        else if (empty($other("to-angle"))) then false()
        else $this("to-angle")=$other("to-angle")
      )
    case "edge" return
      point:same($this("u"),$other("u")) and
      point:same($this("v"),$other("v"))
    case "quad" return
      point:same($this("u"),$other("u")) and
      point:same($this("v"),$other("v")) and
      point:same($this("c1"),$other("c1"))
    case "cubic" return
      point:same($this("u"),$other("u")) and
      point:same($this("v"),$other("v")) and
      point:same($this("c1"),$other("c1")) and
      point:same($this("c2"),$other("c2"))
    default return deep-equal($this,$other)
    )
  )
};

(: 
 : mutate() 
 : Run a function over a sequence of edges to produce a new sequence of
 : edges. The function maps points to points.
 : 
 : $regions: input sequence of edges 
 : $mutate: function that takes a point as an argument and returns a new point
 :)
declare function this:mutate(
  $regions as map(xs:string,item()*)*,
  $mutate as function(item()) as map(xs:string,item()*) (: point to point :)
) as map(xs:string,item()*)*
{
  for $edge in $regions 
  return (
    switch (this:kind($edge))
    case "edge" return (
      this:edge(
        $mutate($edge=>this:start()),
        $mutate($edge=>this:end()),
        $edge=>this:weight(),
        $edge
      )
    )
    case "cubic" return (
      let $controls := $mutate($edge=>this:controls())
      return
        this:cubic(
          $mutate($edge=>this:start()),
          $mutate($edge=>this:end()),
          $controls[1],
          $controls[2],
          $edge=>this:weight(),
          $edge
        )
    )
    case "quad" return (
      let $controls := $mutate($edge=>this:controls())
      return
        this:quad(
          $mutate($edge=>this:start()),
          $mutate($edge=>this:end()),
          $controls[1],
          $edge=>this:weight(),
          $edge
        )
    )
    case "arc" return (
      (: Mutated arcs may not work that well: converts to point arc :)
      this:arc(
        $mutate($edge=>this:arc-center()),
        $edge=>this:arc-radius(),
        $mutate($edge=>this:start()),
        $mutate($edge=>this:end()),
        $edge=>this:arc-flipped(),
        $edge=>this:arc-large()
      )=>
      this:with-properties($edge)
    )
    default return $edge
  )
};

(:
 : reverse()
 : Reverse the direction of the edge.
 :)
declare function this:reverse($edge as map(xs:string,item()*)) as map(xs:string,item()*)
{
  switch(this:kind($edge))
  case "edge" return
    this:edge(
      this:end($edge),
      this:start($edge),
      this:weight($edge),
      $edge
    )
  case "quad" return
    this:quad(
      this:end($edge),
      this:start($edge),
      this:controls($edge)[1],
      this:weight($edge),
      $edge
    )
  case "cubic" return
    this:cubic(
      this:end($edge),
      this:start($edge),
      this:controls($edge)[2],
      this:controls($edge)[1],
      this:weight($edge),
      $edge
    )
  case "arc" return (
    let $circle := $edge=>this:arc-circle()
    let $points := $edge=>this:arc-ends()
    let $angles := $edge=>this:arc-angles()
    return (
      if (exists($points)) then (
        this:arc(
          $circle=>ellipse:center(),
          $circle=>ellipse:radius(),
          $points[2],
          $points[1],
          not($edge=>this:arc-flipped()),
          $edge=>this:arc-large()
        )
      ) else (
        this:arc-by-angle(
          $circle=>ellipse:center(),
          $circle=>ellipse:radius(),
          $angles[2],
          $angles[1],
          not($edge=>this:arc-flipped()),
          $edge=>this:arc-large()
        )
      )
    )
  )
  default return $edge
};

(: edge :)
