xquery version "3.1";
(:
 : Ellipses (including circles)
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/geometric/ellipse"; 

import module namespace errors="http://mathling.com/core/errors"
       at "../core/errors.xqy";
import module namespace util="http://mathling.com/core/utilities"
       at "../core/utilities.xqy";
import module namespace point="http://mathling.com/geometric/point"
       at "point.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

declare variable $this:RESERVED as xs:string* := ("kind", "center", "rx", "ry", "rotation");

(:
 : ellipse()
 : Construct an ellipse. 
 :
 : $center: center of ellipse
 : $rx: axis in x (before rotation)
 : $ry: axis in y (before rotation)
 : $rotation-degrees: how much to rotate ellipse around center
 : $properties: additional properties
 :)
declare function this:ellipse(
  $center as map(xs:string,item()*),
  $rx as xs:double,
  $ry as xs:double,
  $rotation-degrees as xs:double,
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  util:merge-into($properties,
    map {
      "kind": "ellipse",
      "center": $center,
      "rx": $rx,
      "ry": $ry,
      "rotation": $rotation-degrees
    }
  )
};

declare function this:ellipse(
  $center as map(xs:string,item()*),
  $rx as xs:double,
  $ry as xs:double,
  $rotation-degrees as xs:double
) as map(xs:string,item()*)
{
  map {
    "kind": "ellipse",
    "center": $center,
    "rx": $rx,
    "ry": $ry,
    "rotation": $rotation-degrees
  }
};

declare function this:ellipse(
  $center as map(xs:string,item()*),
  $rx as xs:double,
  $ry as xs:double
) as map(xs:string,item()*)
{
  this:ellipse($center, $rx, $ry, 0)
};

(: rx=ry: a circle, in fact :)
declare function this:ellipse(
  $center as map(xs:string,item()*),
  $r as xs:double
) as map(xs:string,item()*)
{
  this:ellipse($center, $r, $r, 0)
};

declare function this:circle(
  $center as map(xs:string,item()*),
  $r as xs:double
) as map(xs:string,item()*)
{
  this:ellipse($center, $r, $r, 0)
};

declare function this:circle(
  $center as map(xs:string,item()*),
  $r as xs:double,
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:ellipse($center, $r, $r, 0, $properties)
};

declare function this:kind($ellipse as map(xs:string,item()*)) as xs:string
{
  $ellipse("kind")
};

declare function this:is-circle($circle as map(xs:string,item()*)) as xs:boolean
{
  switch (this:kind($circle))
  case "ellipse" return (this:rx($circle)=this:ry($circle))
  default return errors:error("GEOM-BADREGION", $circle)
};

declare function this:center(
  $ellipse as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  switch (this:kind($ellipse))
  case "ellipse" return $ellipse("center")
  default return errors:error("GEOM-BADREGION", $ellipse)
};

declare function this:radius($ellipse as map(xs:string,item()*)) as xs:double
{
  if (this:is-circle($ellipse))
  then $ellipse("rx")
  else errors:error("GEOM-BADREGION", $ellipse)
};

declare function this:rx($ellipse as map(xs:string,item()*)) as xs:double
{
  switch (this:kind($ellipse))
  case "ellipse" return $ellipse("rx")
  default return errors:error("GEOM-BADREGION", $ellipse)
};

declare function this:ry($ellipse as map(xs:string,item()*)) as xs:double
{
  switch (this:kind($ellipse))
  case "ellipse" return $ellipse("ry")
  default return errors:error("GEOM-BADREGION", $ellipse)
};

declare function this:rotation($ellipse as map(xs:string,item()*)) as xs:double
{
  switch (this:kind($ellipse))
  case "ellipse" return $ellipse("rotation")
  default return errors:error("GEOM-BADREGION", $ellipse)
};

(:======================================================================
 : Property management
 :======================================================================:)
 
(:
 : property-map()
 : Return the annotation properties of the point as a map. Check whether this
 : is actually an ellipse.
 :
 : $region: the region
 :)
declare function this:property-map(
  $region as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  switch (this:kind($region))
  case "ellipse" return util:exclude($region, $this:RESERVED)
  default return map {}
};

(:
 : properties()
 : Return the names of the annotation properties of the ellipse.
 : Check whether this is actually an ellipse.
 : 
 : $region: the region
 :)
declare function this:properties(
  $region as map(xs:string,item()*)
) as xs:string*
{
  switch (this:kind($region))
  case "ellipse" return ($region=>map:keys())[not(. = $this:RESERVED)]
  default return ()
};

(:
 : with-properties()
 : Annotate the ellipse with some new properties and return the new ellipse.
 : Will not touch any of the core properties. Will override existing properties
 : with the same keys but leave properties with different keys in place.
 : Raises an error if this is not actually an ellipse.
 : 
 : $region: the region
 :)
declare function this:with-properties(
  $region as map(xs:string,item()*),
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  switch (this:kind($region))
  case "ellipse" return 
    util:merge-into($region, util:exclude($properties,$this:RESERVED))
  default return errors:error("GEOM-BADREGION", $region)
};

(:======================================================================:
 : Operations
 :======================================================================:)
 
(:
 : snap()
 : Snap the coordinates of the center, returning the ellipse with snapped
 : (i.e. integer) center coordinates and radii.
 :
 : $ellipses: the ellipses to snap
 :)
declare function this:snap(
  $ellipses as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  for $ellipse in $ellipses return switch(this:kind($ellipse))
  case "ellipse" return
    this:ellipse(
      point:snap(this:center($ellipse)),
      round(this:rx($ellipse)) cast as xs:integer,
      round(this:ry($ellipse)) cast as xs:integer,
      this:rotation($ellipse),
      this:property-map($ellipse)
    )
  default return $ellipse
};

(:
 : decimal()
 : Perform decimal rounding on all the point coordinates (see util:decimal).
 : and radii.
 :
 : $ellipses: the ellipses to round
 : $digits: how many digits after the decimal point to keep
 :)
declare function this:decimal(
  $ellipses as map(xs:string,item()*)*,
  $digits as xs:integer
) as map(xs:string,item()*)*
{
  for $ellipse in $ellipses return switch(this:kind($ellipse))
  case "ellipse" return
    this:ellipse(
      point:decimal(this:center($ellipse), $digits),
      util:decimal(this:rx($ellipse), $digits),
      util:decimal(this:ry($ellipse), $digits),
      this:rotation($ellipse),
      this:property-map($ellipse)
    )
  default return $ellipse
};

(:
 : quote()
 : Return a string value for the ellipses, suitable for debugging.
 : 
 : $ellipses: the ellipses to quote
 :)
declare function this:quote(
  $ellipses as map(xs:string,item()*)*
) as xs:string
{
  string-join(
    for $ellipse in $ellipses return switch(this:kind($ellipse))
    case "ellipse" return
      if (this:is-circle($ellipse)) then (
        "@"||$ellipse("rx")||" "||point:quote($ellipse("center"))
      ) else (
        "@@"||($ellipse("rx"))||"x"||($ellipse("ry"))||" "||point:quote($ellipse("center"))||"∠"||$ellipse("rotation")
      )
    default return errors:quote($ellipse)
    ,
    " "
  )
};

(:
 : same()
 : Equality comparison for ellipses, ignoring annotation properties.
 : Return true() if they have equal coordinates.
 : 
 : $this: one ellipse
 : $other: the ellipse to compare it to
 :)
declare function this:same(
  $this as map(xs:string,item()*),
  $other as map(xs:string,item()*)
) as xs:boolean
{
  let $this-kind := this:kind($this)
  let $other-kind := this:kind($other)
  return
  (
    $this-kind=$other-kind and (
      switch($this-kind)
      case "ellipse" return
        point:same($this("center"),$other("center")) and
        (this:rx($this)=this:rx($other)) and
        (this:ry($this)=this:ry($other)) and
        (this:rotation($this)=this:rotation($other))
    default return deep-equal($this,$other)
    )
  )
};

(: 
 : mutate() 
 : Run a function over a sequence of ellipses to produce a new sequence of
 : ellipses. The function maps points to points.
 : 
 : $regions: input sequence of ellipses 
 : $mutate: function that takes a point as an argument and returns a new point
 :)
declare function this:mutate(
  $regions as map(xs:string,item()*)*,
  $mutate as function(item()) as map(xs:string,item()*) (: point to point :)
) as map(xs:string,item()*)*
{
  for $region in $regions return switch(this:kind($region))
  case "ellipse" return
    let $center := $mutate(this:center($region))
    return
      this:ellipse($center, this:rx($region), this:ry($region),
                   this:rotation($region), this:property-map($region))
  default return $region
};

declare function this:contains-point(
 $ellipse as map(xs:string,item()*),
 $point as map(xs:string,item()*),
 $tolerance as xs:double
) as xs:boolean
{
  if (this:is-circle($ellipse)) then (
    point:distance(this:center($ellipse), $point) <= this:radius($ellipse) + $tolerance
  ) else (
    let $α := this:rotation($ellipse)
    let $a := this:rx($ellipse)
    let $b := this:ry($ellipse)
    let $xp := point:px($point)
    let $yp := point:py($point)
    let $x0 := point:px(this:center($ellipse))
    let $y0 := point:py(this:center($ellipse))
    return (
      ((math:cos($α)*($xp - $x0) + math:sin($α)*($yp - $y0)) div ($a*$a)) +
      ((math:sin($α)*($xp - $x0) + math:cos($α)*($yp - $y0)) div ($b*$b))
      <= 1 + $tolerance
    )
  )
};

declare function this:contains-point(
 $ellipse as map(xs:string,item()*),
 $point as map(xs:string,item()*)
) as xs:boolean
{
  if (this:is-circle($ellipse)) then (
    point:distance(this:center($ellipse), $point) <= this:radius($ellipse)
  ) else (
    let $α := util:radians(this:rotation($ellipse))
    let $a := this:rx($ellipse)
    let $b := this:ry($ellipse)
    let $xp := point:px($point)
    let $yp := point:py($point)
    let $x0 := point:px(this:center($ellipse))
    let $y0 := point:py(this:center($ellipse))
    let $f1 := math:cos($α)*($xp - $x0) + math:sin($α)*($yp - $y0)
    let $f2 := math:sin($α)*($xp - $x0) + math:cos($α)*($yp - $y0)
    return (
      ( ($f1*$f1) div ($a*$a) + ($f2*$f2) div ($b*$b) ) <= 1
    )
  )
};

(:
 : perimeter()
 : Perimeter of the ellipse. Approximately.
 : circle= 2πr
 : ellipse≈ 2π√(a²+b²)
 :)
declare function this:perimeter($ellipse as map(xs:string,item()*)) as xs:double
{
  if (this:is-circle($ellipse))
  then 2*math:pi()*this:radius($ellipse)
  else 2*math:pi()*math:sqrt(this:rx($ellipse)*this:rx($ellipse) + this:ry($ellipse)*this:ry($ellipse))
};

(:
 : area()
 : Area of the ellipse. 
 : circle= πr²
 : ellipse= πab
 :)
declare function this:area($ellipse as map(xs:string,item()*)) as xs:double
{
  math:pi()*this:rx($ellipse)*this:ry($ellipse)
};

(: ellipse :)
