xquery version "3.1";
(:
 : Point objects; 2D and 3D and 4D
 :
 : Points have two kinds of accessors: "precision" accessors that return the
 : raw double coordinates, and basic accessors which snap the coordinates
 : to integer values.
 :
 : Points are maps of coordinates which may be decorated with other 
 : properties (e.g. for rendering). There are also methods for dealing with
 : the point coordinates as sequences: these exist mainly for efficiency in 
 : certain cases (see noise modules and core/vector).
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/geometric/point"; 

import module namespace util="http://mathling.com/core/utilities"
       at "../core/utilities.xqy";
import module namespace errors="http://mathling.com/core/errors"
       at "../core/errors.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

declare variable $this:RESERVED := ("kind", "x", "y", "z", "w", "dim");

(:======================================================================
 : Constructors
 :======================================================================:)

(:
 : point()
 : Construct a 2D point
 :
 : $x: x coordinate
 : $y: y coordinate
 :)
declare function this:point($x as xs:double, $y as xs:double) as map(xs:string,item()*)
{
  map {
    "kind": "point",
    "x": $x,
    "y": $y,
    "dim": 2
  }
};

(:
 : point()
 : Construct a 3D point
 :
 : $x: x coordinate
 : $y: y coordinate
 : $z: z coordinate
 :)
declare function this:point(
  $x as xs:double,
  $y as xs:double,
  $z as xs:double
) as map(xs:string, item())
{
  map {
    "kind": "point",
    "x": $x,
    "y": $y,
    "z": $z,
    "dim": 3
  }
};

(:
 : point()
 : Construct a 4D point
 :
 : $x: x coordinate
 : $y: y coordinate
 : $z: z coordinate
 : $w: w coordinate
 :)
declare function this:point(
  $x as xs:double,
  $y as xs:double,
  $z as xs:double,
  $w as xs:double
) as map(xs:string, item())
{
  map {
    "kind": "point",
    "x": $x,
    "y": $y,
    "z": $z,
    "w": $w,
    "dim": 4
  }
};

(:
 : vector()
 : Construct a point of some dimension from a series of coordinates.
 : Useful for genericizing some of the noise code. It is an error if the
 : number of coordinates is not 2, 3, or 4. Coordinates are given in the
 : (x, y, z w) order.
 :
 : $coordinates: the coordinates
 :)
declare function this:vector(
  $coordinates as xs:double*
) as map(xs:string,item()*)
{
  switch (count($coordinates))
  case 2 return this:point($coordinates[1], $coordinates[2])
  case 3 return this:point($coordinates[1], $coordinates[2], $coordinates[3])
  case 4 return this:point($coordinates[1], $coordinates[2], $coordinates[3], $coordinates[4])
  default return errors:error("GEOM-UNKDIM", (count($coordinates), $coordinates))
};

(:
 : as-dimension()
 : Reconstitute a point as a point with the given number of dimensions.
 : If we are reducing dimensions, just remove excess coordinates. 
 : If we are expanding dimensions, fill in with 0.
 :
 : Examples: 
 :   point:as-dimension(point:point(1,2), 3) == point:point(1,2,0)
 :   point:as-dimension(point:point(1,2,3), 3) == point:point(1,2)
 :
 : $point: source point
 : $dim: dimension to cast point to
 :)
declare function this:as-dimension(
  $point as map(xs:string,item()*),
  $dim as xs:integer
) as map(xs:string,item()*)
{
  let $v := this:vector(this:pcoordinates($point, $dim))
  return
    if (empty(this:properties($point))) then $v
    else util:merge-into($point, $v)
};

(:======================================================================
 : Accessors
 :======================================================================:)

(:
 : dimension()
 : Return the dimension of the point.
 :
 : $point: the point
 :)
declare function this:dimension($point as map(xs:string,item()*)) as xs:integer
{
  (
    $point("dim"),
    if ($point=>map:contains("w")) then 4
    else if ($point=>map:contains("z")) then 3
    else if ($point=>map:contains("y")) then 2
    else 0 (: Dunno :)
  )[1]
};

(:
 : max-dimension()
 : Convenience method to return the maximum dimension of a set of points.
 : 
 : $points: sequence of points
 :)
declare function this:max-dimension($points as map(xs:string,item()*)*) as xs:integer
{
  max(for $pt in $points return this:dimension($pt))
};

(: 
 : x()
 : Accessor for the snapped x coordinate
 :
 : $point: the point
 :)
declare function this:x($point as map(xs:string,item()*)) as xs:integer
{
  round($point("x")) cast as xs:integer
};

(: 
 : y()
 : Accessor for the snapped y coordinate
 :
 : $point: the point
 :)
declare function this:y($point as map(xs:string,item()*)) as xs:integer
{
  round($point("y")) cast as xs:integer
};

(: 
 : z()
 : Accessor for the snapped z coordinate; return 0 for 2D point
 :
 : $point: the point
 :)
declare function this:z($point as map(xs:string,item()*)) as xs:integer
{
  round(head(($point("z"), 0))) cast as xs:integer
};

(: 
 : w()
 : Accessor for the snapped w coordinate; return 0 for 2D or 3D point
 :
 : $point: the point
 :)
declare function this:w($point as map(xs:string,item()*)) as xs:integer
{
  round(head(($point("w"), 0))) cast as xs:integer
};

(: 
 : px()
 : Accessor for the raw precision x coordinate
 :
 : $point: the point
 :)
declare function this:px($point as map(xs:string,item()*)) as xs:double
{
  $point("x")
};

(: 
 : py()
 : Accessor for the raw precision y coordinate
 :
 : $point: the point
 :)
declare function this:py($point as map(xs:string,item()*)) as xs:double
{
  $point("y")
};

(: 
 : pz()
 : Accessor for the raw precision z coordinate; returns 0 for 2D point
 :
 : $point: the point
 :)
declare function this:pz($point as map(xs:string,item()*)) as xs:double
{
  head(($point("z"), 0))
};

(: 
 : pw()
 : Accessor for the raw precision w coordinate; returns 0 for 2D or 3D point
 :
 : $point: the point
 :)
declare function this:pw($point as map(xs:string,item()*)) as xs:double
{
  head(($point("w"), 0))
};

(:
 : coordinates()
 : Return snapped coordinates of the point as a sequence.
 : 
 : $point: the point
 :)
declare function this:coordinates($point as map(xs:string,item()*)) as xs:integer*
{
  let $d := this:dimension($point)
  return (
    if ($d < 2 or $d > 4)
    then errors:error("GEOM-UNKDIM", ($d, $point))
    else (),
    if ($d >= 2) then (this:x($point), this:y($point)) else (),
    if ($d >= 3) then this:z($point) else (),
    if ($d >= 4) then this:w($point) else ()
  )
};

(:
 : coordinates()
 : Return snapped coordinates of the point as a sequence, treating the
 : point as having a given dimension. 
 : 
 : $point: the point
 :)
declare function this:coordinates($point as map(xs:string,item()*), $d as xs:integer) as xs:integer*
{
  if ($d < 2 or $d > 4)
  then errors:error("GEOM-UNKDIM", ($d, $point))
  else (),
  if ($d >= 2) then (this:x($point), this:y($point)) else (),
  if ($d >= 3) then this:z($point) else (),
  if ($d >= 4) then this:w($point) else ()
};

(:
 : pcoordinates()
 : Return raw precision coordinates of the point as a sequence.
 : 
 : $point: the point
 :)
declare function this:pcoordinates($point as map(xs:string,item()*)) as xs:double*
{
  let $d := this:dimension($point)
  return (
    if ($d < 2 or $d > 4)
    then errors:error("GEOM-UNKDIM", ($d, $point))
    else (),
    if ($d >= 2) then (this:px($point), this:py($point)) else (),
    if ($d >= 3) then this:pz($point) else (),
    if ($d >= 4) then this:pw($point) else ()
  )
};

(:
 : pcoordinates()
 : Return raw precision coordinates of the point as a sequence, treating the
 : point as having a given dimension. 
 : 
 : $point: the point
 :)
declare function this:pcoordinates($point as map(xs:string,item()*), $d as xs:integer) as xs:double*
{
  if ($d < 2 or $d > 4)
  then errors:error("GEOM-UNKDIM", ($d, $point))
  else (),
  if ($d >= 2) then (this:px($point), this:py($point)) else (),
  if ($d >= 3) then this:pz($point) else (),
  if ($d >= 4) then this:pw($point) else ()
};

(: 
 : kind()
 : Accessor for the kind of data this is; for debugging
 :
 : $point: the point
 :)
declare function this:kind($point as map(xs:string,item()*)) as xs:string
{
  $point("kind")
};

(:======================================================================:
 : Property management
 :======================================================================:)
 
(:
 : property-map()
 : Return the annotation properties of the point as a map. Check whether this
 : is actually a point.
 : 
 : $region: the region
 :)
declare function this:property-map($region as map(xs:string,item()*)) as map(xs:string, item())
{
  switch(this:kind($region))
  case "point" return util:exclude($region, $this:RESERVED)
  default return map {}
};

(:
 : properties()
 : Return the names of the annotation properties of the point.
 : Check whether this is actually a point.
 : 
 : $region: the region
 :)
declare function this:properties($region as map(xs:string,item()*)) as xs:string*
{
  switch(this:kind($region))
  case "point" return ($region=>map:keys())[not(. = $this:RESERVED)]
  default return ()
};

(:
 : with-properties()
 : Annotate the point with some new properties and return the new point.
 : Will not touch any of the core properties. Will override existing properties
 : with the same keys but leave properties with different keys in place.
 : Raises an error if this is not actually a point.
 : 
 : $region: the region
 :)
declare function this:with-properties(
  $region as map(xs:string,item()*),
  $properties as map(xs:string, item())
) as map(xs:string, item())
{
  switch(this:kind($region))
  case "point" return
    util:merge-into($region, util:exclude($properties,$this:RESERVED))
  default return errors:error("GEOM-BADREGION", $region)
};

(:======================================================================:
 : Point operations
 :======================================================================:)

(:
 : magnitude()
 : Magnitude of the point as vector from origin.
 : 
 : $v: the point
 :)
declare function this:magnitude($v as map(xs:string,item()*)) as xs:double
{
  math:sqrt(
    this:px($v)*this:px($v) +
    this:py($v)*this:py($v) + 
    this:pz($v)*this:pz($v) +
    this:pw($v)*this:pw($v)
  )
};

(:
 : dot()
 : Compute the fot product of two points
 : 
 : $p1: one point
 : $p2: another point
 :)
declare function this:dot(
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*)
) as xs:double
{
  this:px($p1)*this:px($p2) +
  this:py($p1)*this:py($p2) + 
  this:pz($p1)*this:pz($p2) +
  this:pw($p1)*this:pw($p2)
};

(:
 : determinant()
 : Compute the determinant of two 2D points.
 :
 : $u: one point
 : $v: another point
 :)
declare function this:determinant(
  $u as map(xs:string,item()*),
  $v as map(xs:string,item()*)
) as xs:double
{
  $u=>this:px()*$v=>this:py() - $u=>this:py()*$v=>this:px()
};

(:
 : cross()
 : Compute the cross product of two 3D points.
 :
 : $u: one point
 : $v: another point
 :)
declare function this:cross(
  $u as map(xs:string,item()*),
  $v as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:point(
    $u=>this:py()*$v=>this:pz() - $u=>this:pz()*$v=>this:py(),
    $u=>this:pz()*$v=>this:px() - $u=>this:px()*$v=>this:pz(),
    $u=>this:px()*$v=>this:py() - $u=>this:py()*$v=>this:px()
  )
};

(:
 : times()
 : Scale every coordinate by a constant, returning the new point.
 :
 : $u: point to scale
 : $k: constant multiplier
 :)
declare function this:times(
  $u as map(xs:string,item()*),
  $k as xs:double
) as map(xs:string,item()*)
{
  this:map(function ($c) {$k * $c}, $u)
};

(:
 : sum()
 : Add two points, returning the new point.
 : 
 : $u: one point
 : $v: the other point
 :)
declare function this:add(
  $u as map(xs:string,item()*),
  $v as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:map2(function($a,$b) {$a + $b}, $u, $v)
};

(:
 : sub()
 : Subtract one point from another, returning the new point.
 : 
 : $u: one point
 : $v: the other point
 :)
declare function this:sub(
  $u as map(xs:string,item()*),
  $v as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:map2(function($a,$b) {$a - $b}, $u, $v)
};

(:
 : minus()
 : Negate the coordinates of the point, returning the new point.
 : 
 : $v: the point
 :)
declare function this:minus($v as map(xs:string,item()*))
  as map(xs:string,item()*)
{
  this:map(function($a) {-$a}, $v)
};

(:
 : normalize()
 : Normalize a vector represented as a point (scale to unit vector), returning
 : the new point.
 :
 : $v: the point
 :)
declare function this:normalize(
  $v as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  let $l := this:magnitude($v)
  return this:map(function ($c) {$c div $l}, $v)
};

(:
 : perpendicular()
 : Compute the perpendicular vector for a 2D vector represented as a point.
 : (Rotated by 90 degrees.)
 : 
 : $v: the point
 :)
declare function this:perpendicular(
  $v as map(xs:string, item())
) as map(xs:string, item())
{
  this:point($v=>this:py(), -($v=>this:px()))
};

(:
 : perpendiculars()
 : Compute both perpendicular vectors for a 2D vector represented as a point.
 : Rotated by 90 degrees and by -90 degrees, in that order.
 : 
 : $v: the point
 :)
declare function this:perpendiculars($v as item()) as item()*
{
  this:point(this:py($v), -this:px($v)),
  this:point(-this:py($v), this:px($v))
};

(:
 : perpendiculars-with-z()
 : Compute perpendicular vectors for a 3D vector represented as a point.
 : Rotated in y by 90 degrees and by -90 degrees, in that order, with z
 : rotated in the corresponding direction. For spline/smoothing.
 : 
 : $v: the point
 :)
declare function this:perpendiculars-with-z($v as item()) as item()*
{
  this:point(this:py($v), -this:px($v), this:pz($v)),
  this:point(-this:py($v), this:px($v), -this:pz($v))
};

(:
 : snap()
 : Snap the coordinates of the points, returning the points with snapped
 : (i.e. integer) coordinates.
 :
 : $points: the points
 :)
declare function this:snap(
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  for $point in $points return (
    this:map(function ($c) {round($c) cast as xs:integer}, $point)
  )
};

(:
 : floor()
 : Return a point where all the coordinates have been set to the floor of
 : the original point's values. Differs from snap() for negative coordinates.
 :
 : $points: the points
 :)
declare function this:floor(
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  for $point in $points return (
    this:map(fn:floor#1, $point)
  )
};

(:
 : decimal()
 : Perform decimal rounding on all the coordinates (see util:decimal).
 :
 : $points: the points to round
 : $digits: how many digits after the decimal point to keep
 :)
declare function this:decimal(
  $points as map(xs:string,item()*)*,
  $digits as xs:integer
) as map(xs:string,item()*)*
{
  for $point in $points return (
    this:map(function($c) {util:decimal($c, $digits)}, $point)
  )
};

(:
 : quote()
 : Return a string value for the points, suitable for debugging.
 : 
 : $points: the point sequence to quote
 :)
declare function this:quote($points as map(xs:string,item()*)*) as xs:string
{
  string-join(
    for $point in $points return (
      string-join(this:pcoordinates($point)!string(.),",")
    )
    ,
    " "
  )
};

(:
 : same()
 : Equality comparison for points, ignoring annotation properties.
 : Return true() if they have equal coordinates.
 : 
 : $this: one point
 : $other: the point to compare it to
 :)
declare function this:same(
  $this as map(xs:string,item()*),
  $other as map(xs:string,item()*)
) as xs:boolean
{
  this:px($this)=this:px($other) and
  this:py($this)=this:py($other) and
  this:pz($this)=this:pz($other) and
  this:pw($this)=this:pw($other)
};

(: 
 : mutate() 
 : Run a function over a sequence of points to produce a new sequence of
 : points. 
 : 
 : $regions: input sequence of points 
 : $mutate: function that takes a point as an argument and returns a new point
 :)
declare function this:mutate(
  $regions as map(xs:string,item()*)*,
  $mutate as function(map(xs:string,item()*)) as map(xs:string,item()*)
) as map(xs:string,item()*)* {
 $mutate($regions)
};

(:
 : distinct()
 : Return the points in the sequence that have distinct coordinates from
 : each other.
 : 
 : $points: the sequence of points
 :)
declare function this:distinct(
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  let $sorted :=
    for $pt in $points
    order by this:px($pt) ascending, this:py($pt) ascending, this:pz($pt) ascending, this:pw($pt) ascending
    return $pt
  return (
    head($sorted), 
    for $pt at $i in tail($sorted)
    where not(this:same($pt, $sorted[$i]))
    return $pt
  )
};

(:
 : unquote()
 : Parse a string representation of a point and produce a point.
 : The string consists of the coordinates in x,y,z,w order with commas and 
 : optional spaces in between.
 : Example "3.4, 1" == point:point(3.4, 1)
 :
 : $string: the input string to parse
 :)
declare function this:unquote(
  $string as xs:string
) as map(xs:string,item()*)
{
  this:vector(tokenize($string, "[ ]*,[ ]*")!xs:double(.))
};

(:
 : map()
 : Map a function over the coordinates of a point, returning the new point.
 : 
 : $f: function from coordinate to coordinate
 : $point: input point
 :)
declare function this:map(
  $f as function(xs:double) as xs:double,
  $point as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  let $v := this:vector(fn:for-each(this:pcoordinates($point), $f))
  return (
    if (empty(this:properties($point))) then $v
    else util:merge-into($point, $v)
  )
};

(:
 : map()
 : Map a function over the coordinates of a point, returning the new point.
 : Treat the point as having a particular dimension.
 : 
 : $f: function from coordinate to coordinate
 : $point: input point
 : $d: the effective dimension of the point
 :)
declare function this:map(
  $f as function(xs:double) as xs:double,
  $point as map(xs:string,item()*),
  $d as xs:integer
) as map(xs:string,item()*)
{
  let $v := this:vector(fn:for-each(this:pcoordinates($point, $d), $f))
  return (
    if (empty(this:properties($point))) then $v
    else util:merge-into($point, $v)
  )
};

(:
 : map2()
 : Map a function over the coordinates of two points, pair by pair, returning
 : the new point. That is, apply function to the x coordinates, then the
 : y coordinates, and so on.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 :)
declare function this:map2(
  $f as function(xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  let $d := this:max-dimension(($p1, $p2))
  return
    this:vector(
      util:zip($f, this:pcoordinates($p1, $d), this:pcoordinates($p2, $d))
    )
};

(:
 : map2()
 : Map a function over the coordinates of two points, pair by pair, returning
 : the new point. That is, apply function to the x coordinates, then the
 : y coordinates, and so on. Treat the point as having a particular dimension.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 : $d: the effective dimension of the points
 :)
declare function this:map2(
  $f as function(xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*),
  $d as xs:integer
) as map(xs:string,item()*)
{
  this:vector(
    util:zip($f, this:pcoordinates($p1, $d), this:pcoordinates($p2, $d))
  )
};


(:
 : map3()
 : Map a function over the coordinates of three points, triple by triple, 
 : returning the new point. That is, apply function to the x coordinates, 
 : then the y coordinates, and so on.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 : $p3: third input point
 :)
declare function this:map3(
  $f as function(xs:double, xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*),
  $p3 as map(xs:string,item()*)
) as map(xs:string, item()*)
{
  let $d := this:max-dimension(($p1, $p2, $p3))
  return
    this:vector(
      util:zip(
        function ($a, $fc) {$fc($a)},
        this:pcoordinates($p3, $d),
        util:zip(
          function($a, $b) { $f($a, $b, ?) },
          this:pcoordinates($p1, $d), this:pcoordinates($p2, $d)
        )
      )
    )
};

(:
 : map3()
 : Map a function over the coordinates of three points, triple by triple, 
 : returning the new point. That is, apply function to the x coordinates, 
 : then the y coordinates, and so on. Treat the point as having a 
 : particular dimension.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 : $p3: third input point
 : $d: the effective dimension of the points
 :)
declare function this:map3(
  $f as function(xs:double, xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*),
  $p3 as map(xs:string,item()*),
  $d as xs:integer
) as map(xs:string, item()*)
{
  this:vector(
    util:zip(
      function ($a, $fc) {$fc($a)},
      this:pcoordinates($p3, $d),
      util:zip(
        function($a, $b) { $f($a, $b, ?) },
        this:pcoordinates($p1, $d), this:pcoordinates($p2, $d)
      )
    )
  )
};

(:
 : map4()
 : Map a function over the coordinates of four points, coordinate by
 : coordinate, returning the new point. That is, apply function to the x
 : coordinates, then the y coordinates, and so on.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 : $p3: third input point
 : $p4: fourth input point
 :)
declare function this:map4(
  $f as function(xs:double, xs:double, xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*),
  $p3 as map(xs:string,item()*),
  $p4 as map(xs:string,item()*)
) as map(xs:string, item()*)
{
  let $dim := this:max-dimension(($p1, $p2, $p3))
  return
    this:vector(
      util:zip(
        function($d, $fd) {$fd($d)},
        this:pcoordinates($p4, $dim),
        util:zip(
          function ($c, $fc) {$fc($c, ?)},
          this:pcoordinates($p3, $dim),
          util:zip(
            function($a, $b) { $f($a, $b, ?, ?) },
            this:pcoordinates($p1, $dim), this:pcoordinates($p2, $dim)
          )
        )
      )
    )
};

(:
 : map4()
 : Map a function over the coordinates of four points, coordinate by
 : coordinate, returning the new point. That is, apply function to the x
 : coordinates, then the y coordinates, and so on. Treat the point as having a 
 : particular dimension.
 : 
 : $f: function from coordinate pairs to coordinate
 : $p1: one input point
 : $p2: another input point
 : $p3: third input point
 : $p4: fourth input point
 : $d: the effective dimension of the points
 :)
declare function this:map4(
  $f as function(xs:double, xs:double, xs:double, xs:double) as xs:double,
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*),
  $p3 as map(xs:string,item()*),
  $p4 as map(xs:string,item()*),
  $dim as xs:integer
) as map(xs:string, item()*)
{
  this:vector(
    util:zip(
      function($d, $fd) {$fd($d)},
      this:pcoordinates($p4, $dim),
      util:zip(
        function ($c, $fc) {$fc($c, ?)},
        this:pcoordinates($p3, $dim),
        util:zip(
          function($a, $b) { $f($a, $b, ?, ?) },
          this:pcoordinates($p1, $dim), this:pcoordinates($p2, $dim)
        )
      )
    )
  )
};

(:
 : distance()
 : Euclidean distance between points
 :)
declare function this:distance(
  $p1 as map(xs:string,item()*),
  $p2 as map(xs:string,item()*)
) as xs:double
{
  let $d := this:max-dimension(($p1, $p2))
  return (
    math:sqrt(
      sum(
        util:zip(
          function($c1 as xs:double, $c2 as xs:double) as xs:double {
            ($c1 - $c2)*($c1 - $c2)
          },
          this:pcoordinates($p1, $d),
          this:pcoordinates($p2, $d)
        )
      )
    )
  )
};

(: point :)
