xquery version "3.1";
(:
 : Rectangle
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/geometric/rectangle"; 

import module namespace errors="http://mathling.com/core/errors"
       at "../core/errors.xqy";
import module namespace util="http://mathling.com/core/utilities"
       at "../core/utilities.xqy";
import module namespace point="http://mathling.com/geometric/point"
       at "point.xqy";
import module namespace space="http://mathling.com/type/space"
       at "../types/space.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace array="http://www.w3.org/2005/xpath-functions/array";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

declare variable $this:RESERVED := ("kind", "min-x", "min-y", "max-x", "max-y");

(:======================================================================
 : Constructors
 :======================================================================:)

(:
 : rectangle()
 : Construct a rectangle in x/y space.
 :   $x-min: Minimum on x-axis
 :   $y-min: Minimim on y-axis
 :   $x-max: Maximum on x-axis
 :   $y-max: Maximim on y-axis
 :)
declare function this:rectangle(
  $x-min as xs:double,
  $y-min as xs:double,
  $x-max as xs:double,
  $y-max as xs:double
) as map(xs:string,item()*)
{
  (: Instead of tossing an error: rearrange the axes to meet the
   : constraints. This is so we don't just toss errors if some fool
   : does something like shear a rectangle :)
  if ($x-min > $x-max) then (
    this:rectangle($x-max, $y-min, $x-min, $y-max)
  ) else if ($y-min > $y-max) then (
    this:rectangle($x-min, $y-max, $x-max, $y-min)
  ) else (
    map {
      "kind": "box",
      "min-x": $x-min,
      "max-x": $x-max,
      "min-y": $y-min,
      "max-y": $y-max
    }
  )
};

declare function this:rectangle(
  $pt-min as map(xs:string,item()*),
  $pt-max as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:rectangle(
    point:px($pt-min), point:py($pt-min), point:px($pt-max), point:py($pt-max)
  )
};

declare function this:to-space(
  $pt-min as item(),
  $pt-max as item(),
  $properties as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  util:merge-into(
    $properties, 
    space:space(
      abs(point:px($pt-max) - point:px($pt-min)),
      abs(point:py($pt-max) - point:py($pt-min)),
      this:edge($properties)
    )
  )
};

(:======================================================================
 : Accessors
 :======================================================================:)

(:
 : min-point()
 : Accessor for minimal point
 :   $box: rectangle
 :)
declare function this:min-point($box as map(xs:string,item()*)) as map(xs:string,item()*)
{
  point:point(this:min-px($box), this:min-py($box))
};

(:
 : max-point
 : Accessor for maximal point
 :   $box: rectangle
 :)
declare function this:max-point($box as map(xs:string,item()*)) as map(xs:string,item()*)
{
  point:point(this:max-px($box), this:max-py($box))
};

(:
 : min-px()
 : Accessor for raw minimum x-axis of point. 
 :   $box: rectangle
 :)
declare function this:min-px($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return $box("min-x")
  case "space" return 0
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : max-px()
 : Accessor for raw maximum x-axis of point. 
 :   $box: rectangle
 :)
declare function this:max-px($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return $box("max-x")
  case "space" return $box("width")
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : min-py()
 : Accessor for raw minimum y-axis of point. 
 :   $box: rectangle
 :)
declare function this:min-py($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return $box("min-y")
  case "space" return 0
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : max-py()
 : Accessor for raw maximum y-axis of point. 
 :   $box: rectangle
 :)
declare function this:max-py($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return $box("max-y")
  case "space" return $box("height")
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : min-x()
 : Accessor for snapped minimum x-axis of point. 
 :   $box: rectangle
 :)
declare function this:min-x($box as map(xs:string,item()*)) as xs:integer
{
  switch(this:kind($box))
  case "box" return round($box("min-x")) cast as xs:integer
  case "space" return 0
  default return errors:error("GEOM-BADREGION", $box)

};

(:
 : min-x()
 : Accessor for snapped maximum x-axis of point. 
 :   $box: rectangle
 :)
declare function this:max-x($box as map(xs:string,item()*)) as xs:integer
{
  switch(this:kind($box))
  case "box" return round($box("max-x")) cast as xs:integer
  case "space" return round($box("width")) cast as xs:integer
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : min-y()
 : Accessor for snapped minimum y-axis of point. 
 :   $box: rectangle
 :)
declare function this:min-y($box as map(xs:string,item()*)) as xs:integer
{
  switch(this:kind($box))
  case "box" return round($box("min-y")) cast as xs:integer
  case "space" return 0
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : max-y()
 : Accessor for snapped maximum y-axis of point. 
 :   $box: rectangle
 :)
declare function this:max-y($box as map(xs:string,item()*)) as xs:integer
{
  switch(this:kind($box))
  case "box" return round($box("max-y")) cast as xs:integer
  case "space" return 0
  default return errors:error("GEOM-BADREGION", $box)
};

(: 
 : kind()
 : Accessor for the kind of data this is; for debugging
 :
 : $box: the rectangle
 :)
declare function this:kind($box as map(xs:string,item()*)) as xs:string
{
  $box("kind")
};

(:
 : width()
 : Width of rectangle
 :   $box: rectangle
 :)
declare function this:width($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return this:max-px($box) - this:min-px($box)
  case "space" return space:width($box)
  default return errors:error("GEOM-BADREGION", $box)
};

(:
 : height()
 : Height of rectangle
 :   $box: rectangle
 :)
declare function this:height($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return this:max-py($box) - this:min-py($box)
  case "space" return space:width($box)
  default return errors:error("GEOM-BADREGION", $box)
};

declare function this:edge($box as map(xs:string,item()*)) as xs:double
{
  switch(this:kind($box))
  case "box" return 0
  case "space" return ($box=>map:get("edge"),0)[1]
  default return errors:error("GEOM-BADREGION", $box)
};

(:======================================================================:
 : Property management
 :======================================================================:)
 
(:
 : property-map()
 : Return the annotation properties of the box as a map. Check whether this
 : is actually a box.
 : 
 : $region: the region
 :)
declare function this:property-map($region as map(xs:string,item()*)) as map(xs:string, item())
{
  switch(this:kind($region))
  case "box" return util:exclude($region, $this:RESERVED)
  case "space" return util:exclude($region, $space:RESERVED)
  default return map {}
};

(:
 : properties()
 : Return the names of the annotation properties of the box as a map. 
 : Check whether this is actually a box.
 : 
 : $region: the region
 :)
declare function this:properties($region as map(xs:string,item()*)) as xs:string*
{
  switch(this:kind($region))
  case "box" return ($region=>map:keys())[not(. = $this:RESERVED)]
  case "space" return ($region=>map:keys())[not(. = $space:RESERVED)]
  default return ()
};

(:
 : with-properties()
 : Annotate the box with some new properties and return the new box.
 : Will not touch any of the core properties. Will override existing properties
 : with the same keys but leave properties with different keys in place.
 : Raises an error if this is not actually a box.
 : 
 : $region: the region
 :)
declare function this:with-properties(
  $region as map(xs:string,item()*),
  $properties as map(xs:string, item())
) as map(xs:string, item())
{
  switch(this:kind($region))
  case "box" return
    util:merge-into($region, util:exclude($properties,$this:RESERVED))
  case "space" return
    util:merge-into($region, util:exclude($properties,$space:RESERVED))
  default return errors:error("GEOM-BADREGION", $region)
};


(:======================================================================:
 : Box operations
 :======================================================================:)

declare function this:snap(
  $rectangles as map(xs:string,item()*)*
) as map(xs:string,item()*)*
{
  for $rectangle in $rectangles return switch(this:kind($rectangle))
  case "box" return
    this:rectangle(
      this:min-x($rectangle), this:min-y($rectangle),
      this:max-x($rectangle), this:max-y($rectangle)
    )
  case "space" return
    util:merge-into($rectangle, 
      space:space(
        round(space:width($rectangle)) cast as xs:integer,
        round(space:height($rectangle)) cast as xs:integer,
        round(space:edge($rectangle)) cast as xs:integer
      )
    )
  default return $rectangle
};

declare function this:decimal(
  $rectangles as map(xs:string,item()*)*,
  $digits as xs:integer
) as map(xs:string,item()*)*
{
  for $rectangle in $rectangles return switch(this:kind($rectangle))
  case "box" return
    this:rectangle(
      util:decimal(this:min-px($rectangle), $digits),
      util:decimal(this:min-py($rectangle), $digits),
      util:decimal(this:max-px($rectangle), $digits),
      util:decimal(this:max-py($rectangle), $digits)
    )
  case "space" return
    util:merge-into($rectangle,
      space:space(
        util:decimal(space:width($rectangle), $digits),
        util:decimal(space:height($rectangle), $digits),
        util:decimal(space:edge($rectangle), $digits)
      )
    )
  default return $rectangle
};

declare function this:quote(
  $rectangles as map(xs:string,item()*)*
) as xs:string
{
  string-join(
    for $rectangle in $rectangles return switch(this:kind($rectangle))
    case "box" return
      "["||this:min-x($rectangle)||","||this:min-y($rectangle)||","||
           this:max-x($rectangle)||","||this:max-y($rectangle)||"]"
    case "space" return
      "["||this:min-x($rectangle)||","||this:min-y($rectangle)||","||
           this:max-x($rectangle)||","||this:max-y($rectangle)||"]"
    default return errors:quote($rectangle)
    ,
    " "
  )
};

(:
 : same()
 : Equality comparison for rectangles, ignoring annotation properties.
 : Return true() if they have equal coordinates.
 : 
 : $this: one rectangle
 : $other: the rectangle to compare it to
 :)
declare function this:same(
  $this as map(xs:string,item()*),
  $other as map(xs:string,item()*)
) as xs:boolean
{
  this:min-px($this)=this:min-px($other) and
  this:min-py($this)=this:min-py($other) and
  this:max-px($this)=this:max-px($other) and
  this:max-py($this)=this:max-py($other)
};

declare function this:mutate(
  $regions as map(xs:string,item()*)*,
  $mutate as function(map(xs:string,item()*)) as map(xs:string,item()*) (: point to point :)
) as map(xs:string,item()*)*
{
  for $region in $regions return switch(this:kind($region))
  case "box" return
    this:rectangle(
      $mutate(this:min-point($region)),
      $mutate(this:max-point($region))
    )
  case "space" return
    this:to-space(
      $mutate(this:min-point($region)),
      $mutate(this:max-point($region)),
      $region
    )
  default return $region
};

declare function this:contains-point(
 $box as map(xs:string,item()*),
 $point as map(xs:string,item()*),
 $tolerance as xs:double
) as xs:boolean
{
  util:twixt(point:px($point), this:min-px($box) - $tolerance, this:max-px($box) + $tolerance) and
  util:twixt(point:py($point), this:min-py($box) - $tolerance, this:max-py($box) + $tolerance)
};

declare function this:contains-point(
 $box as map(xs:string,item()*),
 $point as map(xs:string,item()*)
) as xs:boolean
{
  util:twixt(point:px($point), this:min-px($box), this:max-px($box)) and
  util:twixt(point:py($point), this:min-py($box), this:max-py($box))
};

(: rectangle :)
