xquery version "3.1";
(:
 : Binary split trees
 : Each node has at most two children containing the points of the parent
 : split along the midpoint of the longest axis. These are 2-dimensional only.
 :
 : Copyright© Mary Holstege 2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/geometric/tree/split"; 

import module namespace sparse="http://mathling.com/core/sparse"
       at "../core/sparse.xqy";
(: import module namespace geom="http://mathling.com/geometric"
       at "euclidean.xqy"; :)
import module namespace point="http://mathling.com/geometric/point"
       at "point.xqy";
import module namespace box="http://mathling.com/geometric/rectangle"
       at "rectangle.xqy";

declare namespace map="http://www.w3.org/2005/xpath-functions/map";

(:
 : split-tree()
 : Construct a 2D split tree from the set of points. The points will
 : be reduced to the set of distinct 2D points.
 :
 : $points: the points to build the tree over
 :)
declare function this:split-tree(
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)
{
  let $tree := this:tree($points)
  return this:naive-split-tree($tree, this:root($tree))
};

(:
 : tree()
 : Initialize the tree from a set of points
 :
 : $points: the set of points
 :)
declare %private function this:tree(
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)
{
  map {
    "kind": "tree",
    "num-nodes": 1,
    "nodes": sparse:array(this:node(0, 0, ()))
  }=>this:put(this:node(1, 1, point:distinct($points!point:as-dimension(.,2))))
};

(:
 : num-nodes()
 : Accessor for the number of nodes in the tree
 :
 : $tree: the split tree
 :)
declare function this:num-nodes(
  $tree as map(xs:string,item()*)
) as xs:integer
{
  $tree("num-nodes")
};

(:
 : node-array()
 : Accessor for the node array in the tree
 :
 : $tree: the split tree
 :)
declare %private function this:node-array(
  $tree as map(xs:string,item()*)
) as map(*)
{
  $tree("nodes")
};

(:
 : nodes()
 : Accessor for the set of nodes in the tree
 :
 : $tree: the split tree
 :)
declare function this:nodes(
  $tree as map(xs:string,item()*)
) as map(xs:string,item()*)*
{
  this:node-array($tree)=>sparse:collection()
};

(:
 : root()
 : Accessor for the root node of the tree
 :
 : $tree: the split tree
 :)
declare function this:root(
  $tree as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  $tree=>this:get(1)
};

(:
 : put()
 : Add/update the node in the tree
 :
 : $tree: the split tree
 : $node: the node to add
 :)
declare function this:put(
  $tree as map(xs:string,item()*),
  $node as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  $tree=>
    map:put("nodes", this:node-array($tree)=>sparse:put(this:id($node), $node))
};

(:
 : get()
 : Get the given node from the tree.
 : 
 : $tree: the split tree
 : $id: the id of the node
 :)
declare function this:get(
  $tree as map(xs:string,item()*),
  $id as xs:integer
) as map(xs:string,item()*)
{
  this:node-array($tree)=>sparse:get($id)
};

(:
 : node()
 : Construct a node in the tree
 : 
 : $id: the ID of the node
 : $depth: depth of the node in the tree
 : $points: set of points for this node
 :)
declare %private function this:node(
  $id as xs:integer,
  $depth as xs:integer,
  $points as map(xs:string,item()*)*
) as map(xs:string,item()*)
{
  map {
    "kind": "node",
    "id": $id,
    "depth": $depth,
    "points": $points,
    "size": count($points)
  }
};

(:
 : id()
 : Accessor for the ID of the node
 :
 : $node: the tree node
 :)
declare function this:id($node as map(xs:string,item()*)) as xs:integer
{
  $node("id")
};

(:
 : depth()
 : Accessor for the depth of the node in the tree
 :
 : $node: the tree node
 :)
declare function this:depth($node as map(xs:string,item()*)) as xs:integer
{
  $node("depth")
};

(:
 : points()
 : Accessor for the set of points belonging to this node
 :
 : $node: the tree node
 :)
declare function this:points(
  $node as map(xs:string,item()*)
) as map(xs:string,item()*)*
{
  $node("points")
};

(:
 : size()
 : Accessor for the number of points belonging to this node
 :
 : $node: the tree node
 :)
declare function this:size(
  $node as map(xs:string,item()*)
) as xs:integer
{
  $node("size")
};

(:
 : box()
 : Accessor for the bounding box of the points belonging to this node.
 :
 : $node: the tree node
 :)
declare function this:box(
  $node as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  $node("box")
};

(:
 : diameter()
 : Diameter of circle containing all the points belonging to this node:
 : the diagonal length of the bounding box.
 :
 : $node: the tree node
 : $metric: the metric to use for distance calculations
 :)
declare function this:diameter(
  $node as map(xs:string,item()*),
  $metric as function(item(),item()) as xs:double
) as xs:double
{
  if (this:size($node) <= 1) then 0
  else (
    let $box := this:box($node)
    return $metric(box:min-point($box), box:max-point($box))
  )
};

(:
 : diameter()
 : Diameter of circle containing all the points belonging to this node:
 : the diagonal length of the bounding box.
 :
 : $node: the tree node
 :)
declare function this:diameter(
  $node as map(xs:string,item()*)
) as xs:double
{
  if (this:size($node) <= 1) then 0
  else (
    this:diameter($node, point:distance#2)
  )
};

(:
 : left()
 : Accessor for the ID of the left child of the node.
 :
 : $node: the tree node
 :)
declare function this:left(
  $node as map(xs:string,item()*)
) as xs:integer
{
  ($node("left"),0)[1]
};

(:
 : right()
 : Accessor for the ID of the right child of the node.
 :
 : $node: the tree node
 :)
declare function this:right(
  $node as map(xs:string,item()*)
) as xs:integer
{
  ($node("right"),0)[1]
};

(:
 : left()
 : The left child of the node.
 :
 : $tree: the split tree
 : $node: the tree node
 :)
declare function this:left(
  $tree as map(xs:string,item()*),
  $node as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:get($tree, this:left($node))
};

(:
 : right()
 : The right child of the node.
 :
 : $tree: the split tree
 : $node: the tree node
 :)
declare function this:right(
  $tree as map(xs:string,item()*),
  $node as map(xs:string,item()*)
) as map(xs:string,item()*)
{
  this:get($tree, this:right($node))
};

(:
 : min-distance()
 : Minimum distance between the points in the two nodes
 : 
 : $node, $other: the two nodes
 : $metric: the distance metric to use
 :)
declare function this:min-distance(
  $node as map(xs:string,item()*),
  $other as map(xs:string,item()*),
  $metric as function(item(),item()) as xs:double
) as xs:double
{
  min(
    for $p in this:points($node)
    for $q in this:points($other)
    return $metric($p, $q)
  )
};


(:
 : min-distance()
 : Minimum distance between the points in the two nodes
 : 
 : $node, $other: the two nodes
 :)
declare function this:min-distance(
  $node as map(xs:string,item()*),
  $other as map(xs:string,item()*)
) as xs:double
{
  if (this:id($node)=this:id($other)) then 0
  (: else if (geom:region-intersects(this:box($node), this:box($other))) then 0 :)
  else this:min-distance($node, $other, point:distance#2)
};

(:
 : naive-split-tree()
 : Recursive naive O(n²) algorithm for splitting the tree.
 :
 : $tree: tree
 : $nodes: set of nodes to split
 :)
declare %private function this:naive-split-tree(
  $tree as map(xs:string,item()*),
  $nodes as map(xs:string,item()*)*
) as map(xs:string,item()*) (: the tree :)
{
  fold-left($nodes, $tree,
    function($tree, $node) {
      switch (this:size($node))
      case 0 return (
        $tree
      )
      case 1 return (
        let $point := this:points($node)
        return 
          $tree=>
            this:put($node=>map:put("box", box:rectangle($point, $point)))
      )
      default return (
        let $depth := this:depth($node)
        let $points := this:points($node)
        let $box :=
          box:rectangle(
            min($points!point:px(.)),
            min($points!point:py(.)),
            max($points!point:px(.)),
            max($points!point:py(.))
          )
        let $d := if (box:width($box) > box:height($box)) then 1 else 2
        let $split :=
          if ($d = 1)
          then avg((point:px(box:min-point($box)), point:px(box:max-point($box))))
          else avg((point:py(box:min-point($box)), point:py(box:max-point($box))))
        let $bottom :=
          if ($d = 1) then (
            $points[point:px(.) < $split]
          ) else (
            $points[point:py(.) < $split]
          )
        let $top := 
          if ($d = 1) then (
            $points[point:px(.) >= $split]
          ) else (
            $points[point:py(.) >= $split]
          )
        let $last-id := this:num-nodes($tree)
        let $left := this:node($last-id + 1, $depth + 1, $bottom)
        let $right := this:node($last-id + 2, $depth + 1, $top)
        let $node :=
          $node=>
            map:put("box", $box)=>
            map:put("left", $last-id + 1)=>
            map:put("right", $last-id + 2)
        let $tree :=
          $tree=>
            map:put("num-nodes", $last-id + 2)=>
            this:put($left)=>
            this:put($right)=>
            this:put($node)
        return
          $tree=>
            this:naive-split-tree(($left, $right))
      )
    }
  )
};

(: splittree :)
