xquery version "3.1";
(:
 : Constructors and accessors for random distributions.
 : Parameters are usually numeric but can be functions to create dynamically
 : mutating distributions. See core/random.xqy for details. 
 :
 : Copyright© Mary Holstege 2020-2021
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :)
module namespace this="http://mathling.com/type/distribution"; 

import module namespace util="http://mathling.com/core/utilities"
  at "../core/utilities.xqy";
import module namespace rand="http://mathling.com/core/random"
  at "../core/random.xqy";
  
declare namespace map="http://www.w3.org/2005/xpath-functions/map";
declare namespace math="http://www.w3.org/2005/xpath-functions/math";

(:======================================================================
 : Basic distribution constructors
 :======================================================================:)

(: 
 : constant()
 : Construct a constant distribution with the given value
 :
 : $value: the value of the constant distribution
 :)
declare %this:distribution function this:constant(
  $value as item()
) as map(xs:string,item()*)
{
  map {
    "distribution" : "constant",
    "min" : $value
  }=>this:set-is-complex()
};

(: 
 : uniform()
 : Construct a uniform distribution over the given value range
 :
 : $min: minimum value of the distribution
 : $max: maximum value of the distribution
 :)
declare %this:distribution function this:uniform(
  $min as item(),
  $max as item()
) as map(xs:string, item()*)
{
  map {
    "distribution" : "uniform",
    "min" : $min,
    "max" : $max
  }=>this:set-is-complex()
};

(:
 : uniform-index()
 : Construct a uniform distribution that functions as an index over
 : integers starting
 :
 : $start: starting index
 : $count: how many index values
 :)
declare %this:distribution function this:uniform-index(
  $start as xs:integer,
  $count as xs:integer
) as map(xs:string, item()*)
{
  map {
    "distribution" : "uniform",
    "min" : $start,
    "max" : $start + $count
  }=>this:set-is-complex()
};


(:
 : uniform-index-of()
 : Construct a uniform distribution that functions as an index over
 : the values
 :
 : $values: the values
 : $count: how many index values
 :)
declare %this:distribution function this:uniform-index-of(
  $values as item()*
) as map(xs:string, item()*)
{
  map {
    "distribution" : "uniform",
    "min" : 1,
    "max" : count($values)
  }=>this:set-is-complex()
};

(: 
 : normal()
 : Construct a normal distribution 
 :
 : $mean: mean value of the distribution
 : $std: standard deviation of the distribution
 :)
declare %this:distribution function this:normal(
  $mean as item(),
  $std as item()
) as map(xs:string, item()*)
{
  map {
    "distribution" : "normal",
    "mean" : $mean,
    "std" : $std
  }=>this:set-is-complex()
};

(: 
 : normal()
 : Construct a normal distribution with default mean (0) and std (1)
 :)
declare %this:distribution function this:normal() as map(xs:string, item()*)
{
  this:normal(0, 1)
};

(: 
 : skewed()
 : Construct a skewed normal distribution 
 :
 : $mean: mean value of the distribution
 : $std: standard deviation of the distribution
 : $skew: skew of the distribution
 :
 : Skewed distribution is defined by parameters α, ξ, and ω
 : α is skew 
 :   α = 0 => normal, α > 0 right skewed, α < 0 left skewed
 :      right skewed => long tail on right
 : ξ = location = shift along x
 : ω = scaling along y
 :   ω is positive 
 : 
 : mean = ξ + ω*δ*sqrt(2/π) δ=α/sqrt(1+α*α) 
 : std = ω*ω*(1 - 2*δ*δ/π)
 :
 : So:
 : ω = sqrt(std / (1 - 2*δ*δ/π))
 : ξ = mean - ω*δ*sqrt(2/π)
 :)
declare %this:distribution function this:skewed(
  $mean as item(),
  $std as item(),
  $skew as item()
) as map(xs:string, item()*)
{
  map {
    "distribution" : "skewed",
    "mean" : $mean,
    "std" : $std,
    "skew" : $skew
  }=>this:set-is-complex()
};

(: 
 : skewed()
 : Construct a skewed normal distribution with default mean, 
 : standard deviation, and skew
 :)
declare %this:distribution function this:skewed() as map(xs:string, item()*)
{
  this:skewed(0, 1, 0)
};
 
(:
 : bernoulli()
 : Construct a Bernoulli distribution
 :
 : $p: Probability of returning 1
 :)
declare %this:distribution function this:bernoulli(
  $p as item()
) as map(xs:string, item()*)
{
  map {
    "distribution" : "bernoulli",
    "p" : $p
  }=>this:set-is-complex()
};

(:
 : bernoulli()
 : Construct a Bernoulli distribution with default probability
 :)
declare %this:distribution function this:bernoulli() as map(xs:string, item()*)
{
  this:bernoulli(50)
};

(:
 : flip()
 : Construct a Bernoulli distribution rendered as boolean values
 :
 : $p: Probability of returning true()
 :)
declare %this:distribution function this:flip(
  $p as item()
) as map(xs:string, item()*)
{
  map {
    "distribution" : "flip",
    "p" : $p
  }=>this:set-is-complex()
};

(:
 : flip()
 : Construct a Bernoulli distribution rendered as boolean values with
 : default probability
 :)
declare %this:distribution function this:flip() as map(xs:string, item()*)
{
  this:flip(50)
};


(:
 : zipf()
 : Construct a Zipf distribution as an integer index from 1 to N
 :
 : p(n)=(1/k^α)/sum(i=1 to n)(1/i^α)
 :    n = number of elements
 :    k = rank
 :    α = exponent
 :
 : $alpha: the α parameter, should be >=1
 :   For English words this by some estimates is around 1.6
 :   For city sizes 1.07
 : $n: number of Zipf values in range
 :)
declare %this:distribution function this:zipf(
  $alpha as xs:double,
  $n as xs:integer
) as map(xs:string, item()*)
{
  let $zipf-weights as xs:double* := rand:zipf-sums($alpha,$n+1)
  return
    map {
      "distribution" : "zipf",
      "cast" : "integer",
      "alpha" : $alpha,
      "sums" : $zipf-weights,
      "max" : $n,
      "truncation" : "resample"
    }=>this:set-is-complex()
};

(:
 : zipf()
 : Construct a Zipf distribution as an integer index from 1 to N with
 : default alpha and limit
 :)
declare %this:distribution function this:zipf() as map(xs:string, item()*)
{
  this:zipf(1,1000)
};

(:
 : markov()
 : Construct a Markov distribution given a square matrix of probabilities
 : expressed as percentages (0 to 100).
 :
 : Example:
 : (0, 10, 90
 :  10, 0, 90,
 :  33, 33, 33)
 : Says state 1 transitions to state 2 10% of the time and state 3 90% of the
 : time; state 2 transitions to state 1 10% of the time and state 3 90% of the
 : time; state3 transitions to each state 1/3 of the time (actually, due to
 : rounding, state 3 transitions to state 3 34% of the time)
 :
 : $start: Starting state (an index in the range 1 to sqrt(count($matrix))
 : $matrix: Square matrix of percentages
 :)
declare %this:distribution function this:markov(
  $start as xs:integer,
  $percent-matrix as xs:integer*
) as map(xs:string, item()*)
{
  let $dim := math:sqrt(count($percent-matrix)) cast as xs:integer
  return 
    map {
      "distribution" : "markov",
      "dim" : $dim,
      "sums" : rand:markov-percent-sums($dim, $percent-matrix),
      "start" : $start
    }=>this:set-is-complex()
};

(:
 : markov()
 : Construct a Markov distribution given a square matrix of probabilities
 : expressed as percentages (0 to 100). Use random start.
 :
 : $percent-matrix: Square matrix of percentages
 :)
declare %this:distribution function this:markov(
  $percent-matrix as xs:integer*
) as map(xs:string, item()*)
{
  let $fixed-dim := math:sqrt(count($percent-matrix)) cast as xs:integer
  let $start := rand:uniform(1,$fixed-dim)=>rand:cast("integer")
  return this:markov($start, $percent-matrix)
};

(:
 : sums()
 : Construct an ad hoc index distribution based on a weight map.
 : Assumes that keys are returned in a consistent order from the table
 : 
 : Example:
 :    {"a": 10, "b": 30, "c": 60}
 : If keys are returned in that order, will return 1 10% of the time, 2
 : 30% of the time, and 3 60% of the time.
 :
 : Usage:
 :
 : $weight-table: map of keys to percentages
 :)
declare %this:distribution function this:sums(
  $weight-table as map(xs:string, xs:double)
) as map(xs:string, item()*)
{
  map {
    "distribution" : "sums",
    "cast" : "integer",
    "sums" : rand:percent-sums($weight-table)
  }=>this:set-is-complex()
};

(: 
 : multimodal()
 : Construct a multimodal distribution
 :
 : $distributions: the component distributions
 :)
declare %this:distribution function this:multimodal(
  $distributions as map(xs:string, item()*)*
) as map(xs:string, item()*)
{
  map {
    "distribution" : "multimodal",
    "distributions" : $distributions
  }=>this:set-is-complex()
};

(:======================================================================
 : Modifier functions
 :======================================================================:)

(:
 : cast()
 : Modify the distribution to cast values. Applies after other modifiers.
 :
 : $distribution: the distribution to modify
 : $cast: one of "integer", "decimal", "boolean", or "string"
 :   decimal will cast to a double value with $rand:DECIMAL-DIGITS following
 :   the decimal point
 :)
declare function this:cast(
  $distribution as map(xs:string, item()*),
  $cast as xs:string
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("cast", $cast)=>
    this:set-is-complex()
};

(:
 : min()
 : Modify the distribution to set a minimum value (after pre-multiplier)
 :
 : $distribution: the distribution to modify
 : $min: the minimum value
 :   for distributions (like normal) that may generate results less than
 :   this, the truncation setting will determine how to handle the case
 :)
declare function this:min(
  $distribution as map(xs:string, item()*),
  $min as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("min", $min)=>
    this:set-is-complex()
};

(:
 : max()
 : Modify the distribution to set a maximum value (after pre-multiplier)
 :
 : $distribution: the distribution to modify
 : $max: the maximum value
 :   for distributions (like normal) that may generate results greater than
 :   this, the truncation setting will determine how to handle the case
 :)
declare function this:max(
  $distribution as map(xs:string, item()*),
  $max as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("max", $max)=>
    this:set-is-complex()
};

(:
 : pre-multiplier()
 : Modify the distribution to set a multiplier to apply to the value before
 : applying min/max and other modifications.
 :
 : $distribution: the distribution to modify
 : $pre-multiplier: the multiplier
 :)
declare function this:pre-multiplier(
  $distribution as map(xs:string, item()*),
  $pre-multiplier as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("pre-multiplier", $pre-multiplier)=>
    this:set-is-complex()
};

(:
 : post-multiplier()
 : Modify the distribution to set a multiplier to apply to the value after
 : applying min/max.
 :
 : $distribution: the distribution to modify
 : $post-multiplier: the multiplier
 :)
declare function this:post-multiplier(
  $distribution as map(xs:string, item()*),
  $post-multiplier as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("post-multiplier", $post-multiplier)=>
    this:set-is-complex()
};

(:
 : truncation()
 : Modify the distribution to set a policy for handling values out of the
 : min/max range. 
 :
 : $distribution: the distribution to modify
 : $truncation: the truncation policy
 :    resample: (the default) retry up to $rand:RESAMPLE-LIMIT times 
 :      and then cap the value to minimum or maximum
 :    ceiling: cap the value to minimum or maximum
 :      Faster, but skews the distribution, perhaps badly
 :    drop: drop the value entirely (i.e. return empty sequence)
 :      Fairly useless for most applications that want a definitive number
 :      of random values
 :)
declare function this:truncation(
  $distribution as map(xs:string, item()*),
  $truncation as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("truncation", $truncation)=>
    this:set-is-complex()
};

(:
 : mean()
 : Modify the distribution to set a mean value. (normal, skewed)
 :
 : $distribution: the distribution to modify
 : $mean: mean value
 :)
declare function this:mean(
  $distribution as map(xs:string, item()*),
  $mean as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("mean", $mean)=>
    this:set-is-complex()
};

(:
 : std()
 : Modify the distribution to set a standard deviation value (normal, skewed)
 :
 : $distribution: the distribution to modify
 : $std: standard deviation value
 :)
declare function this:std(
  $distribution as map(xs:string, item()*),
  $std as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("std", $std)=>
    this:set-is-complex()
};

(:
 : skew()
 : Modify the distribution to set a skew value (skewed)
 :
 : $distribution: the distribution to modify
 : $skew: skew value
 :)
declare function this:skew(
  $distribution as map(xs:string, item()*),
  $skew as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("skew", $skew)=>
    this:set-is-complex()
};

(:
 : p()
 : Modify the distribution to set a probability value (bernoulli, flip)
 :
 : $distribution: the distribution to modify
 : $p: the probability as a percent in [0,100]
 :)
declare function this:p(
  $distribution as map(xs:string, item()*),
  $p as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("p", $p)=>
    this:set-is-complex()
};

(:
 : start()
 : Modify the distribution to set a start value (markov)
 :
 : $distribution: the distribution to modify
 : $start: the starting value of the chain
 :)
declare function this:start(
  $distribution as map(xs:string, item()*),
  $start as item()
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("start", $start)=>
    this:set-is-complex()
};

(:
 : is-complex()
 : Performance hint. If you use the constructors and modifiers in this
 : API you don't need this, but if you are constructing from a vanilla map,
 : and you know the distribution is or is not complex, set it so the 
 : randomizer functions don't have to keep recalculating it. 
 :
 : Don't be wrong here: thing will break. 
 :)
declare function this:is-complex(
  $distribution as map(xs:string, item()*),
  $is-complex as xs:boolean
) as map(xs:string, item()*)
{
  $distribution=>
    map:put("is_complex", $is-complex)
};

(:
 : set-is-complex()
 : Another performance hint, but does the looking for you.
 : Look in algorithm descriptor to see if it needs special handling.
 : The constructors and setters in this API do this automatically; you
 : don't need to.
 :)
declare function this:set-is-complex(
  $distribution as map(xs:string, item()*)
) as map(xs:string, item()*)
{
  $distribution=>map:put("is_complex",
    some $k in map:keys($distribution)
    satisfies (
      typeswitch ($distribution($k))
      case xs:QName return true()
      case function(map(xs:string,item()*), item()?) as item()* return true()
      default return false()
    )
  )
};

(:
 : describe()
 : Dump the algorithm map in form suitable for debugging.
 : Function values will have their QNames dumped (if possible)
 :   distribution: Distribution to use, one of "constant", "uniform", "normal",
 :     "skewed", "bernoulli", "flip", "zipf", "markov", "sums", "multimodal"
 :   min: mimumum value (optional)
 :   max: maximum value (optional)
 :   pre-multiplier: multiplier before min/max (optional)
 :   post-multiplier: multiplier after min/max (optional)
 :   cast: cast type (optional)
 :   mean: mean of distribution (normal, skewed) (default=0)
 :   std: standard deviation of distribution (normal, skewed) (default=mean)
 :   skew: skew of distribution (skewed) (default=0)
 :   p: probability (bernoulli, flip) (default=50)
 :   sums: cumulative probability sums (zipf, markov)
 :   alpha: alpha parameter (zipf) (needed if no sums) (default=0)
 :   limit: number of sums (zipf) (needed if no sums) (default=1000)
 :   start: index of starting symbol (markov) (default=uniform[1,dim])
 :   dim: size of each dimension of Markov matrix (markov)
 :   matrix: raw Markov matrix (used if sums not provided, not recommended)
 :   sums: cumulative probability sums (zipf, markov)
 :     Single element, sequence of doubles
 :   matrix: raw Markov matrix (used if sums not provided, not recommended)
 :     Single element, sequence of doubles
 :   distributions: component distributions (multimodal)
 :)
declare function this:describe($dist as map(xs:string,item()*)) as xs:string
{
  "["||$dist('distribution')||" "||
  string-join((
    this:describe("min",$dist),
    this:describe("max",$dist),
    this:describe("pre-multiplier",$dist),
    this:describe("post-multiplier",$dist),
    this:describe("cast",$dist),
    this:describe("mean",$dist),
    this:describe("std",$dist),
    this:describe("skew",$dist),
    this:describe("p",$dist),
    this:describe("alpha",$dist),
    this:describe("limit",$dist),
    this:describe("dim",$dist),
    this:describe("start",$dist),
    this:describe("sums",$dist,true()),
    this:describe("matrix",$dist,true())
    ), " "
  )||
  (if ($dist=>map:contains("distributions")) then "[" else "")||
  string-join(
    for $distribution at $i in $dist("distributions")
    return this:describe($distribution),
    " "
  )||
  (if ($dist=>map:contains("distributions")) then "]" else "")
};

declare variable $this:CRLF as xs:string := "
";

declare %private function this:describe(
  $k as xs:string,
  $distribution as map(xs:string,item()*),
  $break as xs:boolean
) as xs:string
{
  if ($distribution=>map:contains($k)) then (
    let $v := $distribution=>map:get($k)
    return (
      (if ($break) then $this:CRLF else "")||
      $k||":"||(
        typeswitch ($v)
         case function(*) return util:describe-qname(function-name($v))
         case xs:QName return util:describe-qname($v)
         default return string($v)
      )||
      (if ($break) then $this:CRLF else "")
    )
  ) else ()
};

declare %private function this:describe(
  $k as xs:string,
  $distribution as map(xs:string, item()*)
) as xs:string
{
  this:describe($k, $distribution, false())
};

(: distributions.xqy :)
